#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*====================================

         B P E _ S T A T E

  =====================================*/
AlgState ExQuant::bpe_state(CUBE &Recov_act_cubes)
{

  merging_nodes++;  
  Node &n = Tree_path[curr_depth];
 
  bool cond_holds = check_spec_cond(n);
  if (cond_holds)    {
    finish_proc_sat_cubes();
    Assgn_queue.clear();
    int assgn = (n.lval == 0)?-n.var:n.var;
    push_assgn(Assgn_queue,n.right_ecube,assgn);
    bcp_first_node = SPEC_IMPLIED;
    delete_node(curr_depth);
    return(BCP);
  }

  join_dseqs(n); 
  finish_proc_sat_cubes();

 
  CUBE Late_cubes;
  new_late_cubes(n,Recov_act_cubes,Late_cubes);
  if (Late_cubes.size() > 0)   {
    assert(n.right_ecube >= 0);
    make_cubes_red(n,Late_cubes);
  }
    
  if (xvar_check(n)) seq_gen_ecube(n);

  delete_node(curr_depth);  
  if (curr_depth == -1) return(FINISH);
  else    {// switch on 'yvar_mode' if appropriate
    update_var_mode();
    return(BACKTRACKING);
  }
    
  
} /* end of function bpe_state */

/*========================================

  J O I N _ D S E Q S 

  This function is called when the
  algorithm is in the state BPE.
    
  It resolves sequents of left
  and right branches containing
  the branching variable of n

  ASSUMPTIONS: 
  'n' is not a conflict node that is 
  either n.left_cube < 0 or  
  n.right_child_cube < 0
  or both conditions hold


  =====================================*/
void ExQuant::join_dseqs(Node &n)
{
  assert ((n.left_ecube < 0) || (n.right_ecube < 0)); 


  if (n.left_ecube >= 0) 
    join_unsat_sat(n);
  else if (n.right_ecube >= 0)     {
    CUBE Old_protecting_cubes;
    join_sat_unsat(n,Old_protecting_cubes);
    check_for_new_red_cubes(Old_protecting_cubes);
  }
  else // no empty cubes in left or right branches
    join_sat_sat(n);
     


} /* end of function join_dseqs */


/*===========================================

  M A K E _ C U B E S _ R E D

  This function is called to update D-sequents
  of 'Late_cubes' (i.e. derived cubes 
  that have no D-sequent in the left branch)

  ============================================*/
void ExQuant::make_cubes_red(Node &n,CUBE &Late_cubes)
{

  CUBE &C = F[n.right_ecube];
  CUBE NewSeq;
  int right_lit;

  if (n.lval == 0) right_lit = n.var;
  else right_lit = -n.var;
    
  bool ok = quick_res(NewSeq,C,right_lit);
  assert(ok);      

  for (uint i=0; i < Late_cubes.size(); i++)  {
    int cube_ind = Late_cubes[i];
    assert(Cube_info[cube_ind].state == ACTIVE);

    int L = max_cube_recovery_depth(NewSeq);
    set_new_act_dseq(cube_ind,NewSeq,false);	 
    check_red_cube_in(cube_ind,0,L,true);  
    num_fc_dseqs++;
  }

} /* end of function make_cubes_red */

/*======================================

     N E W _ L A T E _ C U B E S 

  ======================================*/
void ExQuant::new_late_cubes(Node &n,CUBE &Recov_act_cubes,
                             CUBE &Late_cubes)
{

 
  int lbranch_size = Formula_size[n.var-1];
  for (uint i=0; i < Recov_act_cubes.size(); i++)   {
    int cube_ind = Recov_act_cubes[i];
    if (Cube_info[cube_ind].signal == 0) continue; // not a signal cube
  
    if (lbranch_size >= cube_ind+1)	{
      if (Cube_info[cube_ind].state == ACTIVE)   { 
        p();
        printf("a non-late cube  %d is active\n",cube_ind);
        exit(1);
      }
      continue;  // this cube is not late
    }
      
    if (Cube_info[cube_ind].state != ACTIVE) continue; // this cube is not active

    Late_cubes.push_back(cube_ind);
  }


} /* end of function new_late_cubes */


/*===========================================

  F I N I S H _ P R O C _ S A T _ C U B E S

  This function takes care of Y-cubes
  that become active after unassigning
  the branching variable. 

  ============================================*/
void ExQuant::finish_proc_sat_cubes()
{

  CUBE &Cubes = New_sat_cubes[curr_depth];
  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].type != Y_CUBE) continue;
    Cube_info[cube_ind].state = ACTIVE;
    //    if (Cube_info[cube_ind].signal)   num_act_sgnl_cubes++;
  }
   
  Cubes.clear();

} /* end of function finish_proc_sat_cubes */
