#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*========================================

     B C P _ C N F L _ S T A T E

 This function is called when BCP results
 in a conflict.

========================================*/
AlgState ExQuant::bcp_cnfl_state(CUBE &Recov_act_cubes)
{
 
  CUBE E;
  int confl_jump_depth;  
  CUBE Unassgn_lits;
 
  Recov_act_cubes.clear();
  
  int sgnl_cube_depth;
  bool clean;
  gen_cnfl_cube(E,confl_jump_depth,sgnl_cube_depth,clean);
 

  if (cnfl_cube_flag == 1)       {
    printf("CONFLICT! F.size() = %d, Generated Cube:  ",(int)F.size());
    print_Assgns_cut2(E);
    printf("sgnl_cube_depth = %d\n",sgnl_cube_depth);
  }
  
  if ( curr_depth <= confl_jump_depth)  {
      p();
      printf("explanation: curr_depth <= confl_jump_depth\n");
      printf("curr_depth = %d, confl_jump_depth = %d\n",curr_depth,confl_jump_depth);
      cout << "Unit_ycubes-> " << Unit_ycubes << endl;
      cout << "E-> " << E << endl;
      print_Assgns_cut(E);
      print_var_mode(0);
      exit(1);
    }

  while (true)    {
      Node &n = Tree_path[curr_depth];   
      if (curr_depth == confl_jump_depth) break;  
      undo_red_cubes(n,Recov_act_cubes); // make non-redundant the cubes whose D-sequents
                                         // used the assignment of node 'n'   
      recover_sat_cubes(Recov_act_cubes); 
      add_unassgn_literal(Unassgn_lits);
      drop_node(curr_depth);
      if (curr_depth == -1) 	{
	add_cube(E,sgnl_cube_depth,clean); //  can happen if the initial form. is unsat
	return(FINISH); 
      }

    } /* while */

  
 
  add_cube(E,sgnl_cube_depth,clean);
  if (cnfl_cube_flag == 1)   
    print_sgnl_status(F.size()-1,true);

    
  
  CUBE Cubes_to_protect;
  add_cubes_to_protect(Cubes_to_protect,Unassgn_lits); 

  // assign empty cube
  
  Node &n = Tree_path[curr_depth];
  if (n.branch == LEFT) {
    assert(n.left_ecube < 0);
    n.left_ecube = F.size()-1;
  }
  else 
    n.right_ecube = F.size()-1;
   
  
  if (yvar_node(curr_depth)) // set the 'yvar_mode' to true 
    yvar_mode = true; // if the current node assigns an y-var.

   
  if (n.branch == LEFT)    {
    update_vars_cubes(n,Recov_act_cubes); 
    init_queue_impl_assgn(n);
    flip_branch(n);
    bcp_first_node = IMPLIED;
    add_array(Cubes_to_protect, Recov_act_cubes);
    recover_prot_cubes(Cubes_to_protect,confl_jump_depth);
    first_yvar_assigned = 0;
    return(BCP);
  }
  else { // right branch
    undo_red_cubes(n,Recov_act_cubes); // make the cubes whose D-sequents used the
                                       // assignment of node 'n' non-redundant
    recover_sat_cubes(Recov_act_cubes);      
    unassign_var(curr_depth);
    add_array(Cubes_to_protect,Recov_act_cubes); 
    recover_prot_cubes(n,Cubes_to_protect,confl_jump_depth);  
    return(BPE);
  }
 
} /* end of function bcp_cnfl_state */

/*======================================

  A D D _ U N A S S G N _ L I T E R A L

  This function adds to 'Unassgn_lits'
  the assignment to the variable
  of 'curr_depth'

 ASSUMPTIONS:
   1) curr_depth >= 0

=======================================*/
void ExQuant::add_unassgn_literal(CUBE &Unassgn_lits)
{
  Node &n = Tree_path[curr_depth];
  int val = Var_state[n.var-1].val;
  assert(val != 2);
  if (val == 0) Unassgn_lits.push_back(-n.var);
  else Unassgn_lits.push_back(n.var);  

} /* end of function add_unassgn_literal */

/*=====================

   A D D _ C U B E

 =====================*/
void ExQuant::add_cube(CUBE &C,int sgnl_cube_depth, bool clean)
{

  int cube_ind = F.size();  
  F.push_back(C); // add C to the formula
  
  // update Flits0 and Flits1

  for (uint i=0; i < C.size(); i++) {
    int lit = C[i];
    if (lit < 0)
      Flits0[-1*lit-1].push_back(cube_ind);
    else
      Flits1[lit-1].push_back(cube_ind);
  }
  
  CubeInfo el;
  el.type = cube_type_check(C);
  if (el.type == Y_CUBE) {
    num_ycubes++;
    if (!clean) dirt_ycubes++;
  }

  
  Clean.push_back(clean);
 

  if ((sgnl_cube_depth <= curr_depth) && (el.type != Y_CUBE))   {
    el.signal = 1;
    curr_num_sgnl_cubes++;
    if (curr_num_sgnl_cubes > max_num_sgnl_cubes)
      max_num_sgnl_cubes = curr_num_sgnl_cubes;
    if (sgnl_cube_depth >= 0)
      New_sgnl_cubes[sgnl_cube_depth].push_back(cube_ind);
  }
  else 
    el.signal = 0;
 
  

  el.depth = sgnl_cube_depth;

  Cause_lit.push_back(0);
 
  el.state = ACTIVE;
  Cube_info.push_back(el);
  
  CUBE E;
  Act_dseqs.push_back(E);
  Sat_vars.push_back(E);
    
  if (Cube_info[cube_ind].signal)  {
    num_act_sgnl_cubes++;
    increase_sgnl_cube_scores(cube_ind);
  }

  DNF DE;
  Left_dseqs.push_back(DE);
  Red_cube_depth.push_back(-1);
  Ping.push_back(E);
  Ped.push_back(E);
  hash_table_cubes.add_elem();
  hash_table_cubes1.add_elem();
  form_init_prot_cubes();  

  Last_cnfl_cubes.push_back(cube_ind);
  
   
  if (extra_stat_flag)
    Num_of_dseqs.push_back(0);


} /* end of function add_cube */


/*========================================

    U P D A T E _ V A R S _ C U B E S

  This function is called 
  when an empty cube
 is derived in the left branch
 by 'bcp_cnfl_state'  procedure. 

 It is different from   'update_seqs_vars_cubes'
 in two aspects:
 a) it does not update 'Left_red_cubes'
 b) it does not immediately looks for protection
    for new active cubes
=========================================*/
void ExQuant::update_vars_cubes(Node &n,CUBE &Cubes_to_protect)
{
  CUBE  Cubes;
  find_red_cubes_to_update(Cubes,n); 
  recover_active_cubes(Cubes,Cubes_to_protect);
  recover_sat_cubes(Cubes_to_protect);

} /* end of function update_vars_cubes */



/*========================================================

     F I N D _ R E D _ C U B E S _ T O _ U P D A T E

    This function is similar to 'form_left_red_cubes'.
    The only difference is that it does not update the
    array 'Left_red_cubes'

    This function  computes the set 'Cubes'
    of cubes that switch their state from
    redundant to non-redundant.

    This function is meant to update the state
    of redundant cubes

========================================================*/
void ExQuant::find_red_cubes_to_update(CUBE &Cubes, Node &n)
{

  assert(n.branch == LEFT);
  vector <CubeMap> *pM;
  int val = Var_state[n.var-1].val;
  if (val == 0) pM = &Dseqs_to_check0;
  else pM = &Dseqs_to_check1;

  int x_ind = n.var - 1;

  CubeMap &Cands = (*pM)[x_ind];
  CubeMap::const_iterator i;
  for (i=Cands.begin(); i != Cands.end(); ++i) {
    int cube_ind = i->first; 
    assert (Cube_info[cube_ind].state != ACTIVE);
    assert(contains_var(Act_dseqs[cube_ind],n.var));
    Cubes.push_back(cube_ind);
  }
 

} /* end of function find_red_cubes_to_update */
