#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"




/*===========================================

  A D D _ U N M A R K E D _ C U B E S

  =============================================*/
void add_unmarked_cubes(CUBE &Cubes0,CUBE &Cubes1)
{

  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;  

  for (uint i=0; i < Cubes1.size(); i++)    {
    int cube_ind = Cubes1[i];
    if (table[cube_ind] == marker) continue;
    table[cube_ind] = marker;
    Cubes0.push_back(cube_ind);
  }

} /* end of function add_unmarked_cubes */

/*===========================================

  A D D _ C U B E S _ T O _ P R O T E C T

  This function rechecks protection for the
  cubes of Cubes_to_protect
  
  Namely, the cubes that have been active
  when a conflict occured but contained
  assigned (and hence unprotected) literals
  are added to 'Cubes_to_protect')

  ============================================*/
void ExQuant::add_cubes_to_protect(CUBE &Cubes_to_protect,CUBE &Unassgn_lits)
{

  hash_table_cubes.change_marker();
  hash_table_cubes.started_using();
  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;  

  // mark the cubes of Cubes_to_protect
  for (uint i=0; i < Cubes_to_protect.size(); i++)
    table[Cubes_to_protect[i]] = marker;

  for (uint i=0; i < Unassgn_lits.size(); i++)   {
    int lit = Unassgn_lits[i];
    if (lit < 0)   add_unmarked_cubes(Cubes_to_protect,Flits0[-lit-1]);
    else add_unmarked_cubes(Cubes_to_protect,Flits1[lit-1]);
  }

  hash_table_cubes.done_using();

} /* end of function add_cubes_to_protect */

/*=============================================

  R E C O V E R _ A C T I V E _ C U B E S

  This function changes the status of cubes
  listed in Cubes. 

  ==============================================*/
void ExQuant::recover_active_cubes(CUBE &Cubes,CUBE &Cubes_to_protect)
{
   
  for (uint i=0; i < Cubes.size(); i++)  {
    int cube_ind = Cubes[i];
    assert (Cube_info[cube_ind].state != ACTIVE);
    check_red_cube_out(cube_ind);
    
    recover_active_cube(cube_ind);
    Cubes_to_protect.push_back(cube_ind);      	 
  } 

 
} /* end of function recover_active_cubes */


/*======================================

  R E C O V E R _ P R O T _ C U B E S

  =======================================*/
void ExQuant::recover_prot_cubes(CUBE &Cubes_to_protect,int confl_jump_depth)
{

  CUBE Red_cubes;
  confirm_active_cubes(Cubes_to_protect,Red_cubes,confl_jump_depth);  
  check_for_new_red_cubes(Red_cubes);

} /* end of function recover_prot_cubes */

/*====================================================

  R E C O V E R _ S A T _ C U B E S


  This function is different from 'recover_sat_cubes(void)'
  in that it does not immediately looks for protecting
  cubes for the recovered active cubes. Instead,
  the latter are added to 'Cubes_to_protect'

  
  ================================================*/
void ExQuant::recover_sat_cubes(CUBE &Cubes_to_protect)
{

  CUBE &Cubes = New_sat_cubes[curr_depth];
  for (uint i=0; i < Cubes.size(); i++)   {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].state == ACTIVE) continue;
    assert(Cube_info[cube_ind].state == SATISFIED);
    recover_active_cube(cube_ind);	
    Cubes_to_protect.push_back(cube_ind);      
  }

  Cubes.clear();
} /* end of function recover_sat_cubes */


/*============================================================

  U N D O _ R E D _ C U B E S


  This function is called before  deleting the node when
  generating a conflict cube.

  It makes non-redundant the cubes that became redundant 
  due to  assignment to n.var.  

  =============================================================*/
void ExQuant::undo_red_cubes(Node &n,CUBE &Cubes_to_protect)
{
  CUBE Cubes;

  if (n.branch == RIGHT)  {
    if (n.lval == 0) 
      find_candidate_red_cubes(Cubes,Dseqs_to_check1[n.var-1]);
    else  // n.lval == 1
      find_candidate_red_cubes(Cubes,Dseqs_to_check0[n.var-1]);
  }
  else // left branch 
    if (n.lval == 0)      
      find_candidate_red_cubes(Cubes,Dseqs_to_check0[n.var-1]);     
    else  // n.lval == 1
      find_candidate_red_cubes(Cubes,Dseqs_to_check1[n.var-1]);


  // make cubes active again
  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    assert (Cube_info[cube_ind].state != ACTIVE);
 
    remove_lit_presence(cube_ind);
    recover_active_cube(cube_ind);	  
    Cubes_to_protect.push_back(cube_ind);
  }


} /* end of function undo_red_cubes */

