#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*=======================================

  C H E C K _ Y V A R S

  This function checks the correctness
  of free variables (i.e. those of Y)

  ASSUMPTIONS:
     this->num_vars is already set
   
 ========================================*/
void ExQuant::check_yvars()
{

  
  CCUBE marked(num_vars);
 
  marked.assign(marked.size(),0);

  for (uint i=0; i < Yvars.size(); i++)    {
    if ((Yvars[i] <= 0) || (Yvars[i] > num_vars))	{
      p();
      printf("wrong Y-variable %d\n",Yvars[i]);
      printf("num_vars = %d\n",num_vars);
      exit(1);
    }
     
    int var_ind = Yvars[i]-1;
    if (marked[var_ind] == 1)	{
      p();
      printf("variable %d occurs more than once\n",var_ind+1);
      exit(1);
    }
    marked[var_ind] = 1;
  }
    

} /* end of function check_yvars */


/*=========================================

   C H E C K _ I N I T _ S G N L _ C U B E S


  Assumptions:
    1) this->Yvars is formed

    2) this->num_vars is set

 Returns 'true', if all signal cubes
 are X-cubes. Otherwise, returns 'false'

 =======================================*/
bool ExQuant::check_init_sgnl_cubes()
{


  CCUBE Buff;

  Buff.assign(num_vars,0);

  for (uint i=0; i < Yvars.size(); i++)
    Buff[Yvars[i]-1] = 1;

  assert(Init_sgnl_cubes.size() >  0);
  for (uint i=0; i < Init_sgnl_cubes.size(); i++)    {
    int cube_ind = Init_sgnl_cubes[i];
    char found = 0;
    assert(cube_ind < (int) F.size());
    CUBE &C = F[cube_ind];   
    for (uint j=0; j < C.size(); j++)	{
      if (Buff[abs(C[j])-1] == 1) continue;
      found = 1;
      break;
    }
    if (!found)	{
      p();
      printf("initial cube signal cube F[%d] is an Y-cube\n",cube_ind);
      for (uint k=0; k < C.size(); k++) {
        if (k > 0) printf(" ");
        printf("%d",C[k]);
      }
      printf(" 0\n");
      return(false);
    }
  }

  return(true);

}  /* end of function check_init_sgnl_cubes*/

/*=======================================

    R E M _ W R O N G _ Y V A R S

 ========================================*/
void ExQuant::rem_wrong_yvars()
{

  SCUBE S;
  int count = 0;
  for (uint i=0; i < Yvars.size();i++)   {
    int var_ind = Yvars[i]-1;
    if (var_ind + 1 > num_vars)   {
      count++;
      continue;
    }
    if (S.find(Yvars[i]) != S.end()) {
      count++;
      continue;
    }
    S.insert(Yvars[i]);
    
    if (count > 0)
      Yvars[i-count] = Yvars[i];
  }


  Yvars.resize(Yvars.size() - count);

} /* end of function rem_wrong_yvars */



/*=========================================

   R E M _ S G N L _ Y C U B E S


  Assumptions:
    1) Array 'Yvars' has
       been formed

    2) The value of num_vars 
       is correctly formed

 =======================================*/
void ExQuant::rem_sgnl_ycubes()
{


  vector <char> Buff;

  Buff.assign(num_vars,0);

  for (uint i=0; i < Yvars.size(); i++)
    Buff[Yvars[i]-1] = 1;

  assert(Init_sgnl_cubes.size() >  0);
 
  int count = 0;
  for (uint i=0; i < Init_sgnl_cubes.size(); i++)   {
    int cube_ind = Init_sgnl_cubes[i];
    char found = 0;

    CUBE &C = F[cube_ind];   
    for (uint j=0; j < C.size(); j++)	{
      if (Buff[abs(C[j])-1] == 1) continue;
      found = 1;
      break;
    }
    if (!found)    {
      count++;
      Init_sgnl_ycubes.push_back(cube_ind);
      continue;
    }
    if (count > 0) 
      Init_sgnl_cubes[i-count] = Init_sgnl_cubes[i];

  }

  Init_sgnl_cubes.resize(Init_sgnl_cubes.size() - count);


}  /* end of function rem_sgnl_ycubes*/
