#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*===================================================

  D E C _ M A K I N G _ S T A T E _ Y V A R

  =====================================================*/
AlgState ExQuant::dec_making_state_yvar()
{

  if (num_assgn_yvars ==  (int) Yvars.size())     {
    switch_into_xvar_mode();
    return(DECISION_MAKING);
  }

  if (unit_cube_assgn_yvar())
    return(BCP);
  
   
 

  if (num_assgn_yvars == (int) Yvars.size())    {
    switch_into_xvar_mode();
    return(DECISION_MAKING);
  }

  int assgn;

  switch (yvar_dec_mode)   {
  case YVAR_MAX_ACT_CUBES:
    assgn = find_yvar_assgn2();
    break;
  case YVAR_RAND:
    assgn = rand_yvar_assgn();
    break;
  case YVAR_TRIV:
    assgn = triv_yvar_assgn();
    break;
  case YVAR_LIT_ACTIVITY3:
  case YVAR_LIT_ACTIVITY5:
    if (num_sgnl_yvars > 0) 	{
      assgn = find_yvar_assgn4();
      if (assgn != 0) break;	
    }
    if (yvar_dec_mode == YVAR_LIT_ACTIVITY3)
      assgn = find_yvar_assgn3();
    else
      assgn = find_yvar_assgn5();
    break;
  default:
    assert(false);
  }

  if (assgn == 0)    {
    p();
    if (Tree_path.size() < 3) print_tree_path();
    printf("num_sgnl_yvars = %d\n",num_sgnl_yvars);
    printf("num_assgn_yvars = %d\n",num_assgn_yvars);
    printf("Yvars.size() = %d\n",(int) Yvars.size());
    cout << "Yvars-> " << Yvars << endl;
    exit(1);
  }

  if (dec_making_flag) {
    if (assgn < 0)  printf("%4smaking  assignment ~y%d\n"," ",-assgn);
    else printf("%4smaking assignment y%d\n"," ",assgn);
  }
  
  Assgn_queue.clear();
  push_assgn(Assgn_queue,-1,assgn);
  bcp_first_node = DEC_LEFT;
 
  return(BCP); 
} /* end of function dec_making_state_yvar */


/*===================================

  F I N D _ Y V A R _ A S S G N 2
        
  ==================================*/
int ExQuant::find_yvar_assgn2()
{

  int max_prod = -1;
  int best_var_ind = -1;

 
  for (uint i=0; i <  Yvars.size (); i++) {
    int var_ind = Yvars[i] - 1;
    if (Var_state[var_ind].val < 2) continue;
    int count0 = Act_sgnl_cubes_cnt0[var_ind];
    int count1 = Act_sgnl_cubes_cnt1[var_ind];
    
    if (max_prod < count0*count1)   {
      max_prod = count0*count1;
      best_var_ind = var_ind;
    }
  }
  

  if (best_var_ind == -1) return(0);
  if (Act_sgnl_cubes_cnt0[best_var_ind] <
      Act_sgnl_cubes_cnt1[best_var_ind])
    return(best_var_ind + 1);
  else 
    return(-(best_var_ind+1));
 
  
} /* end of function find_yvar_assgn2 */


/*=================================================

  U N I T _ C U B E _ A S S G N _ Y V A R

  ================================================*/
bool ExQuant::unit_cube_assgn_yvar()
{
  while (Unit_ycubes.size() > 0) {
    int cube_ind,assgn;
    pop_assgn(Unit_ycubes,cube_ind,assgn);
    if (Var_state[abs(assgn)-1].val < 2) continue;
    create_impl_var_node(cube_ind,assgn);
    return(true);
  }

  int assgn;
  int count = 0;
  for (int i=Last_cnfl_cubes.size()-1; i >= 0; i--) {
    int cube_ind = Last_cnfl_cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].type != Y_CUBE) continue;
    int L = find_length(F[cube_ind],assgn);
    if (L > 1) {
      Last_cnfl_cubes[i] = -1; // mark this element for removal
      count++;
      continue;
    }

    if (L == 1)   {
      create_impl_var_node(cube_ind,assgn);
      if (count > 0)  remove_marked_elems(Last_cnfl_cubes);
      return(true);  
    }  
  }

  if (count > 0) remove_marked_elems(Last_cnfl_cubes);
  return(false);

} /* end of function unit_cube_assgn_yvar */
/*=================================================

  U P D A T E _ L A S T _ C N F L _ C U B E S

  ===================================================*/
void ExQuant::update_last_cnfl_cubes()
{

  int count = 0;
  for (uint i=0; i < Last_cnfl_cubes.size(); i++) {
    int cube_ind = Last_cnfl_cubes[i];
    if (Cube_info[cube_ind].type != Y_CUBE) {
      Last_cnfl_cubes[i] = -1;
      count++;
    }
  }

  if (count > 0)
    remove_marked_elems(Last_cnfl_cubes);

} /* end of function update_last_cnfl_cubes */
/*==============================================

  F O R M _ U N I T _ Y C U B E _ Q U E U E

  ==============================================*/
void ExQuant::form_unit_ycube_queue()
{

  Unit_ycubes.clear();
  int assgn;
  for (uint i=0; i < F.size(); i++) {
    if (Cube_info[i].state != ACTIVE) continue;
    if (Cube_info[i].type != Y_CUBE) continue;
    int L = find_length(F[i],assgn);
    if (L == 1) {
      assert (Var_state[abs(assgn)-1].xvar == 0);
      push_assgn(Unit_ycubes,i,assgn);
    }
	 
  }

} /* end of function form_unit_ycube_queue */
/*====================================

   R A N D _ Y V A R _ A S S G N
        
  ====================================*/
int ExQuant::rand_yvar_assgn()
{

  int ntries;

  ntries = 1 + (lrand48() % (Yvars.size() - num_assgn_yvars));
  int count = 0;
  for (uint i=0; i <  Yvars.size (); i++) {
    int var_ind = Yvars[i] - 1;
    if (Var_state[var_ind].val < 2) continue;
    if (++count == ntries) {
      if (lrand48() % 2 == 0) return(-(var_ind+1));
      else return(var_ind+1);
    }
  }
  assert(false);
} /* end of function rand_yvar_assgn */


/*============================================

  S W I T C H _ I N T O  _ X V A R _ M O D E

  ===========================================*/
void ExQuant::switch_into_xvar_mode()
{
  yvar_mode = false;
  form_unit_xcube_queue();
  update_last_cnfl_cubes();
  
} /* end of function switch_into_xvar_mode */

