#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <algorithm>

using namespace std;

#include "d0s_pqe.hh"
#include "einline.hh"

/*==================================

  R E A D _ D I M A C S 1

  ===================================*/
int read_dimacs1(FILE *fp,ExQuant &ExQ) 
{int nvars,ncubes,c;
  char buf[BUF_SIZE];
  CUBE cube;
  DNF &D = ExQ.F;
  int ret_value=0;
 
  while (true) {
    if (fgets(buf,BUF_SIZE,fp) == (char *) NULL) {
      p();
      cout << "eof is encountered\n";
      exit(1);
    }
      
    switch (buf[0]) {
    case 'c': /* comment */
      continue; 
    case 'p':  /* the first informative line */
      c = sscanf(buf,"p cnf %d %d", &nvars, &ncubes);
      assert((nvars >= 0) && (ncubes >= 0));
      if ((c == EOF) || (c != 2)) {
        p();
        cout << "wrong format " << buf << endl;
        exit(1);
      }
      goto START;
    default:
      {
        p();
        cout << "wrong format " << buf <<  endl;
        exit(1);
      } 
    }
  }
 
  //     Read the cleauses
   
 START:  
  SCUBE s; 
  int count=1;
  int empty_cube=0;

  while (c != EOF) {
    int v;
    c = fscanf(fp,"%d", &v);
    if (c == 1) {
      if (v!=0) {
        if (s.find(-1*v)!=s.end()) { // in the current cube we have already seen the opposite literal
          empty_cube = 1;
          ret_value = 1;
          fprintf(stderr,"cube number %d is empty!\n",count);
          goto read_next;
              
        }
           	
        if (s.find(v)!=s.end()) {
          fprintf(stderr,"cube number %d contains identical literals\n",count);
          ret_value = 1;
          goto read_next;
        }
       
        s.insert(v);
        cube.push_back(v);
           
        if (abs(v) > nvars) {
          p();
          fprintf(stderr,"variable index (%d) exceeds the declared number of variables (%d)\n",v,nvars);
          exit(1);
        }
      }
      else { // symbol 0 (end of cube) is read in
        if (!empty_cube) 
          D.push_back(cube);
        count++;
        cube.clear();
        s.clear();
        empty_cube = 0;
      }
              
    }	 
  read_next:;
  }/* end of the while loop */

  fclose(fp);
  if (D.size() > unsigned(ncubes)) {
    p();
    fprintf(stderr,"the seen number of cubes (%d) exceeds the declared number of cubes (%d)\n",(int) D.size(),ncubes);
    exit(1);
  }

  ExQ.num_vars = nvars;

  return (ret_value);  
}/* end of function read_dimacs1 */

/*=======================

  R E A D _ D I M A C S

  =======================*/
int read_dimacs(char *fname,ExQuant &ExQ) 
{ FILE *fp;
 

  fp = fopen(fname,"r");
  if (fp == NULL) {
    p();
    cout << "cannot open file " << fname << endl;
    exit(1);
  }
   
  return(read_dimacs1(fp,ExQ));
} /* end of function read_dimacs */
