#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"
extern int form_num;


/*========================================

  D I S C A R D _ D S E Q S

  ========================================*/
void ExQuant::discard_dseqs(CUBE &Non_sgnl_cubes)
{ 
 
  for (uint i=0; i < Non_sgnl_cubes.size(); i++)    {
    int cube_ind = Non_sgnl_cubes[i];
    if (Cube_info[cube_ind].state == REDUNDANT) check_red_cube_out(cube_ind);
    set_correct_cube_state(cube_ind);
  }


} /* end of function discard_d_seqs */

/*==================================================

  R E C O V E R _ A C T I V E _  C U B E S 1


  The difference of this function from
  'recover_active_cubes' is that it ignores
  the cubes of 'Cubes_to_exclude'. I.e. it
  does not try to check if these cubes are
  protected.

  This function changes the status of redundant cubes
  listed in Cubes. 

  ===================================================*/
void ExQuant::recover_active_cubes1(CUBE &Cubes,CUBE &Cubes_to_exclude)
{
 
  
  // mark cubes to exclude

  CUBE &table = hash_table_cubes1.table;
  int &marker = hash_table_cubes1.marker;
  hash_table_cubes1.change_marker();
  hash_table_cubes1.started_using();

  for (uint i=0; i < Cubes_to_exclude.size(); i++)
    table[Cubes_to_exclude[i]] = marker;

  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    assert(Cube_info[cube_ind].signal); 
    if (Cube_info[cube_ind].state == ACTIVE)	{
      p();
      printf("gcount = %ld\n",gcount);
      printf("cube number %d is active\n",cube_ind);
      print_cube1(F[cube_ind],true);
      print_sgnl_status(cube_ind);
      print_Assgns_cut2(F[cube_ind]);
      printf("current node var is %d\n",curr_node().var);
      exit(1);
    }
    check_red_cube_out(cube_ind);
    recover_active_cube(cube_ind); // util.cc 
  }

  CUBE Red_cubes;
  
  confirm_active_cubes1(Cubes,Red_cubes,curr_depth);  
  check_for_new_red_cubes1(Red_cubes); 
  hash_table_cubes1.done_using();
  
} /* end of function recover_active_cubes1 */
/*=================================================================

  U P D A T E _ D S E Q S _ C U B E S

  This function is called before the left branch is flipped to right
 
  It recovers active cubes and forms the array of  sequents
  invalidated due to the  branch flip.

  Notice, that this function does not call 'recover_sat_cubes'
  because every cube satisfied by the current left assignment
  is added to  Left_red_cubes[curr_depth] by 'form_left_red_cubes'

  Returns:
  true: if the number of waiting D-sequents is not empty
  (i.e. if new active cubes appeared)
  false: otherwise


  =================================================================*/
bool ExQuant::update_dseqs_cubes(Node &n)
{
  
  recover_sat_media_cubes();
  CUBE Media_cubes;
  form_left_red_cubes(n,Media_cubes);
  discard_dseqs(Media_cubes);
  
  CUBE &Cubes = Left_red_cubes[curr_depth];
  
  if (Cubes.size() == 0) return(false);

  if (n.lval == 0)    recover_active_cubes1(Cubes,Flits0[n.var-1]);
  else recover_active_cubes1(Cubes,Flits1[n.var-1]);
 
  update_sat_ycubes(New_sat_cubes[curr_depth]);

 
  return(true);
} /* end of function update_dseqs_cubes */


/*=================================================

  R E C O V E R _ S A T _ S G N L  _ C U B E S

  This function sets the state of every signal cube
  C   contained in 'New_sat_cubes[curr_depth]'.

  If C contains a literal that cannot be protected at
  the current depth, C is switched  to state REDUNDANT. 
  Otherwise, the new state of C   is ACTIVE

  ==================================================*/
void  ExQuant::recover_sat_sgnl_cubes(CUBE &Cubes)
{
 
  for (uint i=0; i < Cubes.size(); i++)   { 
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].signal == 0) continue;
    if (Cube_info[cube_ind].state != SATISFIED)   {
      p();
      printf("cube %d is not in state SATISFIED\n",cube_ind);
      print_cube_state(cube_ind);
      //	  printf("F[%d]: ",cube_ind);
      //cout << F[cube_ind] << endl;
      print_Assgns_cut(F[cube_ind]);
      print_sgnl_status(cube_ind);
      printf("curr_depth = %d\n",curr_depth);
      printf("gcount = %ld\n",gcount);
      exit(1);
    }
      
    recover_active_cube(cube_ind);	  
      
  }

  CUBE Red_cubes;
  confirm_active_cubes(Cubes,Red_cubes,curr_depth);
  check_for_new_red_cubes(Red_cubes);
 
} /* end of function recover_sat_sgnl_cubes */
/*===============================================

  B A C K T R A C K _ S T A T E _ X V A R

  ================================================*/
AlgState ExQuant::backtrack_state_xvar()
{
  assert(curr_depth >= 0);
  Node &n = Tree_path[curr_depth];

  if (n.branch == LEFT)    {
    bool new_cubes = update_dseqs_cubes(n);
    if (new_cubes)	{
      init_queue_dec_assgn(n);
      flip_branch(n);
      bcp_first_node = DEC_RIGHT;	  
      return(BCP);
    }
    else { // no new cubes
      recover_sat_sgnl_cubes(New_sat_cubes[curr_depth]); // media X-cubes are recovered by 'update_dseqs_cubes'
      New_sat_cubes[curr_depth].clear(); // no Y-cube can be satisfied by assigning an X-variable
      flip_branch(n);
      return(BPE);
    }
  }  
  else { // n.branch == RIGHT, no need to recover satisfied signal cubes
    recover_sat_media_cubes();
    unassign_var(curr_depth);
    return(BPE);
  }

} /* end of function backtrack_state_xvar */

/*===============================================

  B A C K T R A C K _ S T A T E _ Y V A R

  ================================================*/
AlgState ExQuant::backtrack_state_yvar()
{
  
  if (curr_depth < 0)    {
    p();
    printf("curr_depth = %d\n",curr_depth);
    exit(1);
  }

  Node &n = Tree_path[curr_depth];
  if (n.branch == LEFT) {
    if (Var_state[n.var-1].xvar == 0) // we are flipping a Y-variable when 'yvar_mode == false'
      update_var_mode();
    
    bool new_cubes = update_dseqs_cubes(n);
    if (new_cubes)       {
      init_queue_dec_assgn(n);
      flip_branch(n);
      bcp_first_node = DEC_RIGHT;
      first_yvar_assigned = 0;
      return(BCP);
    }
    else { // no new cubes       
      recover_sat_sgnl_cubes(New_sat_cubes[curr_depth]); // media X-cubes are recovered by 'update_seqs_vars_cubes'
      update_sat_ycubes(New_sat_cubes[curr_depth]);
      flip_branch(n);
      return(BPE);
    }
  }
  else { // n.branch == RIGHT, no need to recover satisfied signal cubes
    recover_sat_media_cubes();
    unassign_var(curr_depth);
    return(BPE);
  }

} /* end of function backtrack_state_yvar */
