#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"


/*=================================

  P R I N T _ R E D _ S T A T E
   
  ================================*/
void ExQuant::print_red_state(int var_ind)
{
  if (Var_state[var_ind].xvar)
    if (!red_xvar(var_ind)) // non-redundant
      printf("x%d(-1) ",var_ind+1);
    else //redundant 
      printf("{x%d} ",var_ind+1);
  else // Y-variable
    if (act_sgnl_cubes_cnt(var_ind) > 0) // non-redundant
      printf("y%d(-1) ",var_ind+1);
    else //redundant 
      printf("{y%d} ",var_ind+1);
  
} /* end of function print_red_state */
/*==========================================

  P R I N T _ A S S G N S _ C U T

  ============================================*/
void ExQuant::print_Assgns_cut(int cube_ind)
{
  print_Assgns_cut(F[cube_ind]);

} /* end of function print_Assgns_cut */
/*=============================================

  P R I N T _ A S S G N S _ C U T
  
  ==============================================*/
void  ExQuant::print_Assgns_cut(CUBE &C)
{

  for (uint i=0; i < C.size();i++) {
    int var_ind = abs(C[i])-1;
     
    int val = Var_state[var_ind].val;
    if (val == 2) // unassigned
      print_red_state(var_ind);
    else { // assigned
      int depth = Var_sel_depth[var_ind];
      assert(depth >= 0);
      Node &n = Tree_path[depth];
      assert(n.var == var_ind+1);
      switch (val)  {
      case 0:
        if (Var_state[var_ind].xvar)
          printf("~x%d(%d) ",n.var,Var_sel_depth[n.var-1]);
        else 
          printf("~y%d(%d) ",n.var,Var_sel_depth[n.var-1]);
        break;
      case 1:
      case 2:
        if (Var_state[var_ind].xvar) printf("x");
        else printf("y");
        if (val == 1)
          printf("%d(%d) ",n.var,Var_sel_depth[n.var-1]);
        else  
          printf("%d(*) ",n.var);
        break;
      default:
        assert(false);
      }
    }
  }
  printf("\n");

} /* end of function print_Assgns_cut */
/*=====================

  P R I N T _ L I T

  ====================*/
void ExQuant::print_lit(int lit)
{
  int var_ind = abs(lit)-1;
  if (Var_state[var_ind].xvar)
    if (lit < 0)
      printf("~x%d",-lit);
    else // lit > 0
      printf("x%d",lit);
  else // Y-variable
    if (lit < 0)
      printf("~y%d",-lit);
    else // lit > 0
      printf("y%d",lit);

} /* end of function print_lit */
/*====================================

  P R I N T _ C U B E _ S T A T E 0

  ====================================*/
void ExQuant::print_cube_state0(int cube_ind)
{ 
  if (Cube_info[cube_ind].state == REDUNDANT)
    if (Cause_lit[cube_ind] == 0)
      printf("red(%d,res)",Red_cube_depth[cube_ind]);
    else      {
      printf("red(%d,",Red_cube_depth[cube_ind]);
      print_lit(Cause_lit[cube_ind]);
      printf(")");
    }
  else
    if (Cube_info[cube_ind].state == SATISFIED)    {
      int lit = Cause_lit[cube_ind];
      int var_ind = abs(lit)-1;
      if (Var_state[var_ind].val == 2) 
        printf("sat_U(%d)",Red_cube_depth[cube_ind]);
      else
        if (branch_type(var_ind) == LEFT)
          printf("sat_L(%d)",Red_cube_depth[cube_ind]);
        else 
          printf("sat_R(%d)",Red_cube_depth[cube_ind]);
    }
    else
      if (Cube_info[cube_ind].state == ACTIVE)
	printf("active");
      else  { 
        if (Cube_info[cube_ind].state == CNFL_RED) {
          printf("cnfl_red");
          return;
        }
        p();
        printf("cube %d is in unknown state %d\n",cube_ind,Cube_info[cube_ind].state);
        exit(1);
      }
 
} /* end of function print_cube_state0 */
/*====================================

  P R I N T _ C U B E _ S T A T E

  ======================================*/
void ExQuant::print_cube_state(int cube_ind, bool seq_flag)
{

  if (cube_ind >= (int) F.size()) return;
  printf("cube %d: ",cube_ind); 
  print_cube1(F[cube_ind]);
  printf(" is in state ");
  print_cube_state0(cube_ind);
  printf("\n");
  print_sgnl_status(cube_ind,true,4);
  if (seq_flag)
    if (Cube_info[cube_ind].state == REDUNDANT)
      print_dseq3(cube_ind);

} /* end of function print_cube_state */
