#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*=============================

  P R I N T _ N O D E

  ==============================*/
void ExQuant::print_node(Node &n)
{
  printf("var = %d\n",n.var);
  printf("lval = %d\n",n.lval);
  
  if (n.branch == LEFT) printf("LEFT branch\n");
  else    {
    printf("RIGHT branch\n");
    printf("left branch time is %d\n",Formula_size[n.var-1]);
  }
  printf("left_child_cube = %d\n",n.left_ecube);
  printf("right_ecube = %d\n",n.right_ecube);

} /* end of function print_node */

/*=============================

  P R I N T _ T R E E _ P A T H

  ==============================*/
void ExQuant::print_tree_path()
{

  assert(curr_depth + 1 == (int) Tree_path.size());
  printf("Tree_path.size() = %d\n",(int) Tree_path.size());
  for (int i=0; i <= curr_depth; i++) {
    printf("--------------------------");
    printf("node of depth %d\n",i);
    print_node(Tree_path[i]);
  }

 
} /* end of function print_tree_path */

/*========================================

  P R I N T _ V A R _ M O D E

  =========================================*/
void ExQuant::print_var_mode(int num_spaces)
{
  for (int i=0; i < num_spaces; i++) printf(" ");
  if (yvar_mode) printf("YVAR-MODE\n");
  else printf("XVAR-MODE\n");

} /* end of function print_var_mode */
/*========================================
    
  P R I N T _ A L G _ S T A T E 0
      
  This function prints the algorithm 
  state (a short form)

  =======================================*/
void ExQuant::print_alg_state0(AlgState astate)
{

  switch (astate) {
  case DECISION_MAKING:
    printf("DECISION_MAKING");
    break; 
  case  BCP:
    printf("BCP");
    break;
  case BCP_CNFL:
    printf("BCP_CNFL");
    break;
  case BPE:
    printf("BPE");
    break;
  case BACKTRACKING:
    printf("BACKTRACKING "); 
    break;
  case FINISH:
    printf("FINISH");
    break;
  case RESOURCE_EXCEEDED:
    printf("RESOURCE_EXCEEDED");
    break;
  default:
    p();
    printf("print_alg_state:: unknown state!!\n");
    exit(1);
  }
  printf("\n");

} /* end of function print_alg_state0 */

/*========================================
    
  P R I N T _ A L G _ S T A T E
      
  This function prints the algorithm 
  state (a long form)

  =======================================*/
void ExQuant::print_alg_state(bool flag)
{
  
  
  Node *pn = NULL;
  
  if (curr_depth >= 0) pn = &Tree_path[curr_depth];
  if (flag)  printf("----------------------\n");
  printf("State: ");   
  switch (alg_state)   {
  case DECISION_MAKING:
    if (yvar_mode) printf("Yvar ");
    else printf("Xvar ");
    printf("DECISION_MAKING at depth %d",curr_depth);     
    break; 
  case  BCP:
    printf("BCP at depth %d",curr_depth);
    break;
  case BCP_CNFL:
    printf("BCP_CNFL at depth %d",curr_depth);
    break;
  case BPE:
    assert(curr_depth >= 0);
    printf("BPE");
    if (Var_state[pn->var-1].xvar) 
      printf(" for variable x%d(%d,%d)",pn->var,pn->lval,!pn->lval);
    else 
      printf(" for variable y%d(%d,%d)",pn->var,pn->lval,!pn->lval);
    printf(" at depth %d",curr_depth);
    printf("       (left_ecube = %d, right_ecube = %d)",pn->left_ecube,pn->right_ecube);
    break;
  case BACKTRACKING:
    assert(curr_depth >= 0);
    printf("BACKTRACKING "); 
    if (pn->branch == LEFT)  
      if (Var_state[pn->var-1].xvar)
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~x%d,LEFT) at depth %d",pn->var,curr_depth);
        else
          printf("to node n(x%d,LEFT) at depth %d",pn->var,curr_depth);
      else // Y-variable
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~y%d,LEFT) at depth %d",pn->var,curr_depth);
        else 
          printf("to node n(y%d,LEFT) at depth %d",pn->var,curr_depth);
    else //right branch
      if (Var_state[pn->var-1].xvar) {
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~x%d,RIGHT) at depth %d",pn->var,curr_depth);
        else
          printf("to node n(x%d,RIGHT) at depth %d",pn->var,curr_depth);
      }
      else {
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~y%d,RIGHT) at depth %d",pn->var,curr_depth);
        else 
          printf("to node n(y%d,RIGHT) at depth %d",pn->var,curr_depth);
      }
    break;
  case FINISH:
    printf("FINISH");
    break;
  case RESOURCE_EXCEEDED:
    printf("RESOURCE_EXCEEDED");
    break;
  default:
    p();
    printf("print_alg_state:: unknown state!!\n");
    exit(1);
  }
  printf("\n");

} /* end of function print_alg_state */
