#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh" 
#include "einline.hh"


/*==========================================

  E X T R A C T _ P A I R

  ==========================================*/
void ExQuant::extract_pair(int &cube_ind2,int &lit_ind2,
                           int cube_ind1)
{
  lit_ind2 = Ped[cube_ind1].back();
  Ped[cube_ind1].pop_back();
  cube_ind2 = Ped[cube_ind1].back();
  Ped[cube_ind1].pop_back();

} /* end of function extract_pair */

/*============================================

  F I N D _ N E W _ P R O T E C T I O N

  This function tries to find a new protection
  for every active cube listed in 'Ped[cube_ind1]'
  because F[cube_ind1] just became redundant


  =============================================*/
void ExQuant::find_new_protection(CUBE &Red_cubes,int cube_ind1)
{

  int cube_ind2,lit_ind2;
 
  
  while (Ped[cube_ind1].size() > 0)   {
    extract_pair(cube_ind2,lit_ind2,cube_ind1);
    if (Cube_info[cube_ind2].state != ACTIVE) 
      continue;
         

    int lit = F[cube_ind2][lit_ind2];
    if (Var_state[abs(lit)-1].val < 2)      
      continue; // assigned literals cannot be protected
       
    if (!prot_media_cubes)
      if (Cube_info[cube_ind2].signal == 0) continue; // no need to protect non-signal cube

    if (already_protected(cube_ind2,lit_ind2))       
      continue; // in case 'Ped' contains duplicates
      
     
    bool success = protect_one_cube_lit(cube_ind2,lit_ind2);
    if (success) continue;
  
    CUBE S;
    seq_gen_red_cube1(S,lit);
    num_frw_blocked_cube_dseqs++;

    int L = max_cube_recovery_depth(S);
    set_new_act_dseq(cube_ind2,S,false);
    check_red_cube_in(cube_ind2,lit,L,true);    
   
    Red_cubes.push_back(cube_ind2);
  }

} /* end of function find_new_protection */


/*==============================================

  C H E C K _ F O R _ N E W _ R E D _ C U B E S

  This function updates the status of
  cubes protected by cubes listed in 'Red_cubes'
  that just got redundant.

  If no new protected cube exists for a cube,
  the latter is added to 'Red_cubes'
  
  ==============================================*/
void ExQuant::check_for_new_red_cubes(CUBE &Red_cubes)
{

  while (Red_cubes.size() > 0)    {
    int cube_ind = Red_cubes.back();
    Red_cubes.pop_back();
    find_new_protection(Red_cubes,cube_ind);
  }


} /* end of function check_for_new_red_cubes */



/*=====================================================

  P R O T E C T _ O N E _ C U B E _ L I T

  This function returns 'true' if
  an active cube that can be resolved with F[cube_ind]
  on variable abs(F[cube_ind][lit_ind]) is found

  ======================================================*/
bool ExQuant::protect_one_cube_lit(int cube_ind,int lit_ind)
{
  CUBE &C = F[cube_ind];
  CUBE *pCubes;
  int lit = C[lit_ind];
  if (lit < 0) pCubes = &Flits1[-lit-1];
  else pCubes = &Flits0[lit-1];

  int max_depth;
  int ping_cube_ind = fnd_one_ping_cube(C,*pCubes,max_depth);

  if (ping_cube_ind < 0) return(false);

  // update arrays Ping and Ped
  update_prot_cube_info(cube_ind,lit_ind,ping_cube_ind);
  return(true);

} /* end of function protect_one_cube_lit */


/*========================================

  F N D _ O N E _ P I N G _ C U B E

  This functions returns the index of
  an ACTIVE cube from 'Cube_inds' that
  can be resolved with 'C'.

  If no such cube exists, the function
  returns '-1'. In this case, the function
  also computes the maximum depth up to 
  which cube 'C' remains redundant

  =========================================*/
int  ExQuant::fnd_one_ping_cube(CUBE &C,CUBE &Cube_inds,int &max_depth)
{
 
  max_depth = -1;
  for (uint i=0; i < Cube_inds.size(); i++)   {
    int cube_ind = Cube_inds[i];  
    if (Cube_info[cube_ind].state != ACTIVE) {
      if (Red_cube_depth[cube_ind] > max_depth)
        max_depth = Red_cube_depth[cube_ind];
      continue;
    }
    int dist = distance(C,F[cube_ind]);
    assert(dist > 0);
    if (dist == 1) return(cube_ind);
  }

  return(-1);

} /* end of function find_one_ping_cube */


/*============================================

  U P D A T E _ P R O T _ C U B E _ I N F O

  =============================================*/
void ExQuant::update_prot_cube_info(int cube_ind,int lit_ind,
                                    int ping_cube_ind)
{

  Ping[cube_ind][lit_ind] = ping_cube_ind;
 
  Ped[ping_cube_ind].push_back(cube_ind);
  Ped[ping_cube_ind].push_back(lit_ind);

  if (Ped[ping_cube_ind].size() > 20)
    squeeze_Ped(ping_cube_ind);
    

} /* end of function update_prot_cube_info */

/*===================================================

  F N D _ P I N G _ C U B E S

  This function returns 'true' if every variable
  of cube 'F[cube_ind]' is protected i.e. if
  F[cube_ind] can be resolved on its every
  X-variable  with a non-redundant cube

  Otherwise, it returns 'false' and 'lit' is set
  to the literal of F[cube_ind] that cannot be
  protected

  =====================================================*/
bool ExQuant::fnd_ping_cubes(int cube_ind,int &lit)
{

  CUBE &C = F[cube_ind];
  for (uint i=0; i < C.size(); i++){
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].xvar == 0) continue; // skip Y-variables
    bool success = protect_one_cube_lit(cube_ind,i);
    if (!success)   {
      lit = C[i];
      return(false);
    }	
  }

  return(true);

} /* end of function fnd_ping_cubes */







