#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*========================================

 C L A S S    I N I T _ B C P _ D A T A

=========================================*/
class init_BCP_data
{

public:
  int num_vars;
  CUBE assgn_queue;
  CCUBE Assgns;
  CCUBE Sat_cube;
  CCUBE Yvar;
  DNF Lit_vects0; 
  DNF Lit_vects1; 

  CUBE New_sgnl_cubes;
  CCUBE Sgnl_cube; 
  CUBE New_indexes; 
  bool sgnl_flag;

// methods
  
  BCPreturn run_bcp(DNF &F);
  int  make_one_assignment(DNF &F,CUBE &removal_array,CUBE &sat_array);
  int find_length(CUBE &C,int &assgn);
  void mark_yvars(CUBE &Yvars);
  void fill_up_queue(DNF &F,bool &init_empty_cube);
  void init(DNF &F,CUBE &Init_sgnl_cubes);
  void rem_lits(DNF &F);
  void remove_sat_cubes(DNF &F);
  void form_sgnl_cubes(DNF &F,CUBE &Init_sgnl_cubes);
  bool ycube(CUBE &C);
  void form_result(DNF &Res,DNF &F);
 

}; /* end of  class init_BCP_data */

/*========================================

    P R O C _ U N I T _ X _ C U B E S

  Return 
    0 if the number of signal cubes equals 0
      or F contains an empty cube 
      or an empty cube is derived by BCP
    1 otherwise
 
========================================*/
int ExQuant::proc_unit_X_cubes(DNF &Res)
{

 
  init_BCP_data IBD;
  IBD.num_vars = num_vars;  
 
  IBD.mark_yvars(Yvars);
  init_empty_cube = false;
  IBD.fill_up_queue(F,init_empty_cube);
  if (init_empty_cube)    {
    CUBE E;
    Res.clear();
    Res.push_back(E);
    return(0);
  }
   
  
  IBD.init(F,Init_sgnl_cubes);
  BCPreturn answer = IBD.run_bcp(F);
  if (answer == CONFLICT)   {
    CUBE E;
    Res.clear();
    Res.push_back(E);
    init_empty_cube = true;
    return(0);
  }
  IBD.rem_lits(F);
  IBD.remove_sat_cubes(F);
  IBD.form_sgnl_cubes(F,Init_sgnl_cubes);
  if (Init_sgnl_cubes.size() == 0)   {
    empty_init_sgnl_cubes = true;
    IBD.form_result(Res,F);
    return(0);
  }

  return(1);
} /* end of function proc_unit_X_cubes */

/*===========================

    F O R M _ R E S U L T

  ==========================*/
void init_BCP_data::form_result(DNF &Res, DNF &F)
{

  for (uint i=0; i < F.size(); i++)    {
    if (ycube(F[i]))
      Res.push_back(F[i]);
  }

} /* end of function form_result */
/*======================

      Y C U B E

  ====================*/
bool init_BCP_data::ycube(CUBE &C)
{

  for (uint i=0; i < C.size(); i++)    {
    int var_ind = abs(C[i])-1;
    if (Yvar[var_ind] == 0) return(false); // not a Y-cube
  }
 
  return(true);

} /* end of function ycube */

/*========================================

        F O R M _ S G N L _ C U B E S

=========================================*/
void init_BCP_data::form_sgnl_cubes(DNF &F,CUBE &Init_sgnl_cubes)
{

  Init_sgnl_cubes.clear();
  

  for (uint i=0; i < New_sgnl_cubes.size(); i++)   {
    int cube_ind = New_sgnl_cubes[i];
    if (Sat_cube[cube_ind] == 1) continue;
    if (Sgnl_cube[cube_ind] == 0)   {
      p();
      printf("cube %d is not a signal cube\n",cube_ind);
      exit(1);
    }
    if ((Sgnl_cube[cube_ind] == 1) && !ycube(F[New_indexes[cube_ind]]))	{
      Init_sgnl_cubes.push_back(New_indexes[cube_ind]);
      Sgnl_cube[cube_ind] = 2;
    }
  }

 
} /* end of function form_sgnl_cubes */

/*========================================

        R E M O V E _ S A T _ C U B E S

=========================================*/
void init_BCP_data::remove_sat_cubes(DNF &F)
{

  int count = 0;
  for (uint i=0; i < F.size(); i++)   {
    if (Sat_cube[i]) 	{
      count++;
      continue;
    }
    if (count == 0)       {
      New_indexes[i] = i;
      continue;
    }
    F[i-count].clear();
    F[i-count] = F[i];
    New_indexes[i] = i-count;
  }
  
  F.resize(F.size()-count);
} /* end of function remove_sat_cubes */

/*==========================

    M A R K _ Y V A R S

============================*/
void init_BCP_data::mark_yvars(CUBE &Yvars)
{
  Yvar.assign(num_vars,0);
  for (uint i=0; i < Yvars.size(); i++)
    Yvar[Yvars[i]-1] = 1;
} /* end of function mark_yvars */

/*==================================

     F I L L _ U P _ Q U E U E

===================================*/
void init_BCP_data::fill_up_queue(DNF &F,bool &init_empty_cube)
{
  for (uint i=0; i < F.size(); i++)   {
    if (F[i].size() > 1) continue;
    if (F[i].size() == 0)      {
      init_empty_cube = true;
      return;
    }
    int assgn  = -F[i][0];
    int var_ind = abs(assgn)-1;
    if (Yvar[var_ind]) 
      continue; 
    push_assgn(assgn_queue,i,assgn);
  }
} /* end of function fill_up_queue */

/*=======================

         I N I T 

========================*/
void init_BCP_data::init(DNF &F,CUBE &Init_sgnl_cubes)
{

  Sat_cube.assign(F.size(),0); 
 
  Assgns.assign(num_vars,2);

  for (int i=0; i < num_vars; i++)    {
    CUBE Dummy;
    Lit_vects0.push_back(Dummy);
    Lit_vects1.push_back(Dummy);  
  }

  initialize_cube_vects(F,Lit_vects0,Lit_vects1);

  Sgnl_cube.assign(F.size(),0);

  for (uint i=0; i < Init_sgnl_cubes.size(); i++)  {
    int cube_ind = Init_sgnl_cubes[i];
    New_sgnl_cubes.push_back(cube_ind);
    Sgnl_cube[cube_ind] = 1;
  }

  New_indexes.assign(F.size(),-1);

  
} /* end of function init */

/*====================================
     
         F I N D _ L E N G T H

======================================*/
inline int init_BCP_data::find_length(CUBE &C,int &assgn)
{
  int length = 0;
  for (uint i=0; i < C.size();i++)
    if (Assgns[abs(C[i])-1]==2)   {
      if (length++ == 1) // there are at least two unassigned literals 
        return(2);
      assgn = -C[i];
      continue;
    }
  return(length);	  
} /* end of function find_length */

/*================================================

       M A K E _ O N E _ A S S I G N M E N T

==================================================*/
int  init_BCP_data::make_one_assignment(DNF &F,CUBE &removal_array,CUBE &sat_array)

 {
  int cube_ind;
  int der_assgn;
 
  for (uint i=0; i < removal_array.size();i++)   {
    cube_ind = removal_array[i];
    if (Sat_cube[cube_ind])
      continue; 
    if (sgnl_flag)    {
      Sgnl_cube[cube_ind] = 1;
      New_sgnl_cubes.push_back(cube_ind);
    }

    CUBE &C = F[cube_ind];
    int length = find_length(C,der_assgn);
    
    if (length ==0) 
      return(cube_ind); 

    if (length ==1) 
      if (Yvar[abs(der_assgn)-1] == 0) { 
        assgn_queue.push_back(cube_ind);
        assgn_queue.push_back(der_assgn);  
      }
   }


  for (uint i = 0; i < sat_array.size();i++) {
    cube_ind = sat_array[i];
    if (Sat_cube[cube_ind])
      continue; 
    Sat_cube[cube_ind] = 1;
  }
return(-1);
} /* end of function make_one_assignment */

/*==================================================

        R U N _ B C P

===================================================*/
BCPreturn init_BCP_data::run_bcp(DNF &F)
{

  while (true) {
    if (assgn_queue.size() == 0)
      return(NON_CONFLICT);
      
  
    int unit_cube_num, assgn,unsat_cube;
    pop_assgn(assgn_queue,unit_cube_num,assgn);
 
   
 
    if (Sat_cube[unit_cube_num]) 
       continue;
     
 
   int var_ind = abs(assgn)-1;
  
   if (Sgnl_cube[unit_cube_num])   
       sgnl_flag = true;
   else sgnl_flag = false;
   
   if (assgn < 0) {
      Assgns[var_ind] = 0;
      unsat_cube = make_one_assignment(F,Lit_vects0[var_ind],Lit_vects1[var_ind]);
     }
   else   {
      Assgns[var_ind] = 1;
      unsat_cube = make_one_assignment(F,Lit_vects1[var_ind],Lit_vects0[var_ind]); 
     }
      
   if (unsat_cube < 0)
     continue; 
   else     {  
       assgn_queue.clear();      
       return(CONFLICT);
     }
  }
  
} /* end of procedure run_bcp */


/*========================

    R E M _ L I T S

========================*/
void init_BCP_data::rem_lits(DNF &F)
{

  for (uint i= 0; i < F.size(); i++) {
    if (Sat_cube[i]) continue;
    CUBE &C = F[i];
    int count = 0;
    for (uint j =0; j < C.size(); j++) {
      int var_ind = abs(C[j])-1;
      if (Assgns[var_ind] != 2) {
        count++;
        continue;
      }
      if (count == 0) continue;
      C[j-count] = C[j];
    }
    if (count > 0)  C.resize(C.size()-count);
  }


} /* end of function rem_lits */
