#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*===================================================

  J O I N _ S A T _ U N S A T 

  
  This function is called when the right
  branch of node 'n' is assigned an empty cube 'C'.

  This function resolves  C with every D-sequent of
  Waiting_seqs[curr_depth] on the branching variable x 
  of 'n' and makes the resolvent a new active D-sequent

  ====================================================*/
void ExQuant::join_sat_unsat(Node &n,CUBE  &Old_protecting_cubes)
{

  // 'left_lit' is the literal of n.var in the sequents of 
  // Left_red_cubes[curr_depth]


  hash_table_cubes.change_marker();
  hash_table_cubes.started_using();
  int left_lit; 
  if (n.lval == 0)    {
    left_lit = -n.var;
    mark_sat_cubes(Flits0[n.var-1]); // mark cubes satisfied in the right branch
  }
  else    {
    left_lit = n.var;
    mark_sat_cubes(Flits1[n.var-1]); // mark cubes satisfied in the right branch
  }
    
  CUBE &Red_cubes = Left_red_cubes[curr_depth];
 
  bool ok;
  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;
  CUBE &C = F[n.right_ecube];

 
  bool int_var_flag=false;
  if (Var_state[n.var-1].xvar) int_var_flag = true;
  for (uint i=0; i < Red_cubes.size();i++)   {
    int cube_ind = Red_cubes[i];
    rem_old_left_dseqs(cube_ind,n.var);
   
    if (Cube_info[cube_ind].signal == 0)   {
      if (Cube_info[cube_ind].state == REDUNDANT)   check_red_cube_out(cube_ind);
      set_correct_cube_state(cube_ind);  
      continue;
    }
    CUBE &S = Left_dseqs[cube_ind].back();
 
    CUBE NewSeq;
   
    bool recompute_flag = true;
    num_resR_dseqs++;

    if (table[cube_ind] == marker)	{     
      if (Cube_info[cube_ind].state == REDUNDANT)	 {
        p();
        printf("cube_ind = %d is redundant\n",cube_ind);
        printf("gcount = %ld\n",gcount);
        printf("F.size() = %d\n",(int) F.size());
        print_sgnl_status(cube_ind);
        exit(1);
      }
	  
      ok = quick_res(NewSeq,S,left_lit);
      assert(ok);
        
      num_sc_dseqs++;
      int L = max_cube_recovery_depth(NewSeq);
      set_new_act_dseq(cube_ind,NewSeq,false);
      check_red_cube_in(cube_ind,0,L,recompute_flag);
      finish_update_left_dseqs(cube_ind);
      continue;
    }

    if (!contains_lit(S,left_lit))	{
      p();
      printf("D-sequent for cube %d ",cube_ind); cout << S << endl;
      printf("does not contain literal %d\n",left_lit);
      exit(1);
    }
     
    
    ok = resolve(NewSeq,S,C,left_lit);
    if (int_var_flag) num_res_xvars++;
    else num_res_yvars++;
    assert(ok);

    if (Cube_info[cube_ind].state == REDUNDANT) 	{
      remove_lit_presence(cube_ind); // remove literal presence of the current sequent
      recompute_flag = false;
    }
    else	
      Old_protecting_cubes.push_back(cube_ind);     
    
    num_fc_dseqs++;
    
    int L = max_cube_recovery_depth(NewSeq);
    set_new_act_dseq(cube_ind,NewSeq,false); 
    check_red_cube_in(cube_ind,0,L,recompute_flag); 
    finish_update_left_dseqs(cube_ind);
  }
  Red_cubes.clear(); // empty the corresponding list of waiting seqs
  hash_table_cubes.done_using();

  CUBE Late_cubes;
  
  compute_late_cubes(n,Late_cubes);
  proc_late_cubes(n,Late_cubes);
  if (num_act_sgnl_cubes > 0)
    proc_missed_cubes(n);
 
} /* end of function join_sat_unsat */


/*=====================================

  P R O C _ M I S S E D _ C U B E S

  The current algorithm can miss
  some late cubes that do not have
  a D-sequent at all. This can happen
  if all the backtracking is done
  in the  BCP_CNFL state

  ====================================*/
void ExQuant::proc_missed_cubes(Node &n)
{

  int right_lit; 
  if (n.lval == 0) 
    right_lit = n.var;
  else  // lval == 1
    right_lit = -n.var;
  
  CUBE S;
  int cube_ind = n.right_ecube;
  quick_res(S,F[cube_ind],right_lit);
  
  int count=num_act_sgnl_cubes;
  int last_late_cube = Formula_size[n.var-1];

  for (int i=F.size()-1; i >= last_late_cube; i--)    {
    if (Cube_info[i].state != ACTIVE) continue;
    if (Cube_info[i].signal == 0) continue;
    count--;
    if (i == cube_ind) continue; // D-sequent for cube F[n.right_ecube] is built later
    num_resR_dseqs++;
    num_fc_dseqs++;

    int L = max_cube_recovery_depth(S);
    set_new_act_dseq(i,S,false);
    check_red_cube_in(i,0,L,true);  
    if (count == 0) break;
  }
  
  assert(count == 0);
  
} /* end of function proc_missed_cubes */

/*======================================

  Q U I C K _ R E S

  This function copies to 'S1'
  all the literals of 'S0' but 'lit'.

  It returns 'true' if 'S0' contains
  'lit'. 

  =======================================*/
bool quick_res(CUBE &S1,CUBE &S0,int lit)
{

  bool found = false;
 
  for (uint i=0; i < S0.size(); i++)    {
    if (S0[i] == lit) {
      found = true;
      continue;
    }
    S1.push_back(S0[i]);
  }

  return(found);

} /* end of function quick_res */

/*====================================

  M A R K _ S A T _ C U B E S


  This function is called to mark
  cubes of 'Cubes' in the hash table

  ==================================*/
void ExQuant::mark_sat_cubes(CUBE &Cubes)
{

  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;

  for (uint i=0; i < Cubes.size(); i++)
    table[Cubes[i]] = marker;


} /* end of function mark_sat_cubes */

/*===============================================

  F I N I S H _ U P D A T E _ L E F T _ D S E Q S

  =================================================*/
void ExQuant::finish_update_left_dseqs(int red_cube_ind)
{

  Left_dseqs[red_cube_ind].pop_back();
  Sat_vars[red_cube_ind].pop_back();

} /* end of function finish_update_left_dseqs */

/*============================================

  R E M _ O L D _ L E F T _ D S E Q S

  This function removes old left sequents
  i.e. old sequents (starting from the top)
  that do not contain 'branch_var')

  ===========================================*/
void ExQuant::rem_old_left_dseqs(int red_cube_ind,int branch_var)
{

  DNF &L = Left_dseqs[red_cube_ind];
 
  CUBE &Vars = Sat_vars[red_cube_ind]; 

  while (L.size() > 0)   {
    if (Vars.back() == branch_var) return;
    Vars.pop_back();
    L.pop_back();   
  }
 
  // should not reach this line
  p();
  printf("no left sequent for F[%d] contains var. %d\n",red_cube_ind,branch_var);
  printf("The left sequents for F[%d]: \n",red_cube_ind);
  print_dnf(L);
  printf("Here are corresponding Sat_vars: ");
  cout << Vars << endl;
  exit(1);
} /* end of function rem_old_left_dseqs */
