#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*===============================================

  J O I N _ S A T _ S A T


  This function is called when neither branch
  of node 'n' is assigned an empty cube

  ================================================*/
void ExQuant::join_sat_sat(Node &n)
{

  if (res_dseqs_flag)
    printf("entering res_sequents_no_ecube\n");
  int left_lit; // the literal of n.var in left sequents
 
  
  hash_table_cubes.change_marker();
  hash_table_cubes.started_using();
  if (n.lval == 0)    {
    left_lit = -n.var;
    mark_sat_cubes(Flits0[n.var-1]);  // Mark cubes satisfied in the right branch
  }
  else     {
    left_lit = n.var;
    mark_sat_cubes(Flits1[n.var-1]);  // Mark cubes satisfied in the right branch
  }
  // process the sequents of Waiting_seqs
  CUBE &Red_cubes = Left_red_cubes[curr_depth];
 
  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;
  bool ok;

  bool xvar_flag=false;
  if (Var_state[n.var-1].xvar) xvar_flag = true;
  for (uint i=0; i < Red_cubes.size();i++)  {
    int cube_ind = Red_cubes[i];
    if (Cube_info[cube_ind].signal == 0)       {
      if (Cube_info[cube_ind].state == REDUNDANT)  check_red_cube_out(cube_ind);	  
      set_correct_cube_state(cube_ind);
      continue;
    }
    
    // find left sequent
       
    rem_old_left_dseqs(cube_ind,n.var);
    CUBE NewSeq;
   

    CUBE &Sl = Left_dseqs[cube_ind].back();

    if (table[cube_ind] == marker)      { 
      if (Cube_info[cube_ind].state == REDUNDANT)   {
        if (Red_cube_depth[cube_ind] < curr_depth) continue; // no need to do anything if the current D-sequent does not depend on 'n.var'
        remove_lit_presence(cube_ind);
        Act_dseqs[cube_ind].clear();  
      }
	
         
      ok = quick_res(NewSeq,Sl,left_lit);
      assert(ok);

      int L = max_cube_recovery_depth(NewSeq);
      set_new_act_dseq(cube_ind,NewSeq,false);
      check_red_cube_in(cube_ind,0,L,false);
      finish_update_left_dseqs(cube_ind);
	 
      num_resS_dseqs++;
      num_sc_dseqs++;
      continue;
    }

     
     
    // check the right sequent Sr
     
    CUBE &Sr = Act_dseqs[cube_ind];

     
    if (contains_lit(Sr,-left_lit)) { // Sr contains the literal '-left_lit'
      ok = resolve(NewSeq,Sl,Sr,left_lit);
      assert(ok);
      if (xvar_flag) num_res_xvars++;
      else num_res_yvars++;	
    }
    else {// 'Sr' contains no literal '-left_lit'
      ok = resolve(NewSeq,Sl,Sr,left_lit);
      assert(ok);
      if (xvar_flag) num_res_xvars1++;
      else num_res_yvars1++;        
         
    }

    
  
    int L = max_cube_recovery_depth(NewSeq);
    set_new_act_dseq(cube_ind,NewSeq,true);
    check_red_cube_in(cube_ind,0,L,false);
     
     
    num_resS_dseqs++;
    finish_update_left_dseqs(cube_ind);
  }
 

  Red_cubes.clear(); 
  hash_table_cubes.done_using();

  CUBE Late_cubes;
  compute_late_cubes(n,Late_cubes);
  proc_late_cubes(n,Late_cubes);

} /* end of function join_sat_sat */


/*======================================================

  P R O C _ L A T E _ C U B E S

  This function produces D-sequents of late cubes
  that are satisfied in the left branch of 'n.var' 
  (and so have a trivial D-sequent)

  ========================================================*/
void ExQuant::proc_late_cubes(Node &n,CUBE &Late_cubes)
{

  int left_lit;
  if (n.lval == 0) left_lit = -n.var;
  else  left_lit = n.var;

  int count = 0;
  for (uint i=0; i < Late_cubes.size(); i++)   {
    int cube_ind = Late_cubes[i];
    CUBE &S = Act_dseqs[cube_ind];
    CUBE NewSeq;

     
    count++;
         
    bool ok = quick_res(NewSeq,S,-left_lit);
    if (!ok)	{
      p();
      printf("failed to find literal %d in D-sequent of cube  number %d: ",-left_lit,cube_ind);
      print_dseq3(cube_ind);
      print_cube(cube_ind);
      print_Assgns_cut(S);
      printf("gcount = %ld\n",gcount);
      exit(1);
    }
	  
    int L = max_cube_recovery_depth(NewSeq);	  
    set_new_act_dseq(cube_ind,NewSeq,true);
    check_red_cube_in(cube_ind,0,L,false);  

    
	      
  }

  if (n.right_ecube >= 0) num_resR_dseqs += count;
  else if (n.left_ecube == -1) num_resS_dseqs += count;
  else assert(false); 

} /* end of function proc_late_cubes */

/*================================================

  C O M P U T E _ L A T E _ C U B E S

  On exit, an element of 'Late_cubes' contains the
  index of cube C such that
  
  1) The right D-sequent of C depends on 'n.var'

  2) Formula_size[n.var-1] < index(C)+1

  So, C is derived AFTER the left branch of the node
  n was finished

  =================================================*/
void ExQuant::compute_late_cubes(Node &n,CUBE &Late_cubes)
{
  
  assert(n.left_ecube == -1);
 
  int var_ind  = n.var-1;

  int lbranch_size = Formula_size[var_ind];

  CUBE Cand_cubes;
  if (n.lval == 0)
    find_candidate_red_cubes(Cand_cubes,Dseqs_to_check1[var_ind]);
  else
    find_candidate_red_cubes(Cand_cubes,Dseqs_to_check0[var_ind]);


  for (uint i=0; i < Cand_cubes.size(); i++)   {
    int cube_ind = Cand_cubes[i];
    if (Cube_info[cube_ind].signal == 0)    {
      if (Cube_info[cube_ind].state == REDUNDANT)
        check_red_cube_out(cube_ind);
      set_correct_cube_state(cube_ind);        
      continue;
    }
    if (lbranch_size >= cube_ind+1) continue;
    Late_cubes.push_back(cube_ind);
  }
      

} /* end of function compute_late_cubes */

