#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

using namespace std;
#include "d0s_pqe.hh"


//
//  Global variables
//

hsh_tbl hash_table_lits;
hsh_tbl hash_table_aux;
hsh_tbl hash_table_cubes;
hsh_tbl hash_table_cubes1;




/*======================

    M A I N

=======================*/
int  main(int argc,char *argv[])

{
  ExQuant ExQ;
 

  if (argc == 1) {
    ExQ.print_descr();
    exit(0);
  }
 

 

  char fname[MAX_NAME];
  strcpy(fname,argv[1]);

  // read CNF formula
  strcat(fname,".cnf");
  read_dimacs(fname,ExQ);
  ExQ.init_solver_state();

  // read free (i.e. unquantified) variables
  fname[0] = 0;
  strcpy(fname,argv[1]);
  strcat(fname,".txt");
  read_numbers(fname,ExQ.Yvars); 
     
 
  // read cubes to take out of the formula
  fname[0] = 0;
  strcpy(fname,argv[1]);
  strcat(fname,".clauses");
  read_numbers(fname,ExQ.Init_sgnl_cubes);
  
  // read parameters
  ExQ.read_parameters(argc,argv);
  
  if (ExQ.rigor_check) {
    bool ok = ExQ.check_init_sgnl_cubes();
    if (!ok) exit(1);
    if (ok && ExQ.just_check_sgnl_cubes)
      exit(0);
  }
  else
    ExQ.rem_sgnl_ycubes();
  
  if (ExQ.ign_wrong_yvars) ExQ.rem_wrong_yvars();
  else   ExQ.check_yvars(); // check correctness of Y-variables
  DNF Res1;
  int answer = 1;
  double usrtime0=0.,systime0=0.;
  double usrtime=0.,systime=0.;

  if (!ExQ.prepr_mode || ExQ.prepr_bcp)
    answer = ExQ.proc_unit_X_cubes(Res1);
 
  if (ExQ.prepr_mode && !ExQ.prepr_red_cubes)  {
      answer = 2;
      goto FINISH;
    }
 

  
  get_runtime (usrtime0, systime0);
  srand48((long) ExQ.seed_value);

  
  if (answer != 0) answer = ExQ.ds_pqe();
  get_runtime (usrtime, systime);

 
 FINISH:
  DNF Res;
  fname[0] = 0;
  if (strlen(ExQ.res_root_name) == 0) {
    strcpy(ExQ.res_root_name,argv[1]);
    strcat(ExQ.res_root_name,".res");
  }
  strcpy(fname,ExQ.res_root_name);
  strcat(fname,".cnf");
  switch (answer) {
  case 0: 
      {
        printf("successful termination\n");  
        if (ExQ.init_empty_cube || ExQ.empty_init_sgnl_cubes) {
          print_dnf(Res1,fname,ExQ.num_vars);
        }
        else	    {
          ExQ.form_result(Res);
          print_dnf(Res,fname,ExQ.num_vars);
        }
        break;
      }
  case 1:      
    ExQ.form_result(Res);
    print_dnf(Res,fname,ExQ.num_vars);
    break;
  case 2: // preprocessing
    if (ExQ.num_act_sgnl_cubes == 0)   {
      ExQ.form_result(Res);
      print_dnf(Res,fname,ExQ.num_vars);
      printf("Solved by preprocessing!!\n");
      exit(0);
    }
    if (!ExQ.prepr_red_cubes)
      ExQ.print_prepr_results1();
    else
      ExQ.print_prepr_results2();
    break;
  default: assert(false); // shouldn't reach this line
  }

 
  if (!ExQ.init_empty_cube && !ExQ.empty_init_sgnl_cubes) 
     ExQ.print_info(Res.size());
 
  printf("total time is %.2f sec.\n",usrtime-usrtime0);

  if ((answer == 1) || (answer == 2)) {
    printf("not finished\n");
    exit(1);
   }
  exit(0);
  
  
} /* end of function main */

