#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh" 
#include "einline.hh"



/*=================================================

  A D D _ D S E Q S _ I M P L _ N O D E 

  This function is called when an implied node
  is created.  It generates D-sequents corresponding
  to active cubes that get satisified in the left
  branch of 'n.var'.

  Since the left branch is immediately flipped to
  right, new D-sequents are not added to the set
  of active D-sequents. Instead, the new D-sequents
  are directly added to 'Left_seqs'.

 ==================================================*/
void ExQuant::add_dseqs_impl_node(Node &n)
{

  CUBE S;
  CUBE *pSat_cubes;

  
  if (n.lval == 0)   {
    pSat_cubes = &Flits1[n.var-1];
    S.push_back(-n.var);
  }
  else     {
    pSat_cubes = &Flits0[n.var-1];
    S.push_back(n.var);
  }

  for (uint i=0; i < pSat_cubes->size(); i++)    {
    int cube_ind = (*pSat_cubes)[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].signal == 0) continue;
    Left_dseqs[cube_ind].push_back(S);
    Sat_vars[cube_ind].push_back(n.var);
  }

} /* end of function add_dseqs_impl_node */

/*====================================

      C R E A T E _ N O D E

=====================================*/
void ExQuant::create_node(int unit_cube_num,int assgn)
{

  Node n;
  
  n.var = abs(assgn);
  int val;
  if (assgn < 0) val = 0;
  else val = 1;

  Var_state[n.var-1].val = val;


  if (unit_cube_num < 0)    {
    n.branch = LEFT;
    n.lval = val;
    n.left_ecube = -1;
    n.right_ecube = -1;
    Var_sel_depth[n.var-1] = curr_depth+1;
  }
  else     {
    n.branch = RIGHT;
    n.lval = val ^ 1;
    n.left_ecube = unit_cube_num;
    n.right_ecube = -1;   
    Var_sel_depth[n.var-1] = curr_depth+1;
    add_dseqs_impl_node(n);
    Formula_size[n.var-1] = F.size();
    if (Cube_info[unit_cube_num].signal) 
      add_bcp_sgnl_cubes(assgn,curr_depth+1);
  }
    
  Tree_path.push_back(n);
  curr_depth++;

} /*end of function create_node */

/*===========================================

         D E L E T E _ N O D E

============================================*/
void ExQuant::delete_node(int &curr_depth)
{
 
  revoke_sgnl_status(); 
  Tree_path.pop_back();
  New_sgnl_cubes[curr_depth].clear(); 
  curr_depth--;
  
  
} /* end of function delete_node */

/*=====================================

  U N A S S I G N _ V A R

  This function is called after BCP
  terminates in the state 'CONFLICT'.
  It just unassigns the variable
  of the last node (without any
  backtracking)

  =======================================*/
void ExQuant::unassign_var(int curr_depth)
{

  Node &n = Tree_path[curr_depth];
 
  Var_state[n.var-1].val = 2;

  if (Var_state[n.var-1].xvar) 
    num_assgn_xvars--;
  else
    num_assgn_yvars--;
   
} /* end of function unassign_var */

/*=====================================

  U N A S S I G N _ V A R

  Another version of 'unassign_var'
  (with a different parameter)
  =======================================*/
void ExQuant::unassign_var(Node &n)
{

  Var_state[n.var-1].val = 2;
 
  if (Var_state[n.var-1].xvar)
    num_assgn_xvars--;
  else
    num_assgn_yvars--;
   
} /* end of function unassign_var */


/*=====================

  D R O P  _ N O D E

  ====================*/
void ExQuant::drop_node(int &curr_depth)
{

  unassign_var(curr_depth);
  delete_node(curr_depth);

} /* end of function drop_node */


/*===================================


    M A K E _ N O D E _ A S S G N

  This function is called when
  node n=Tree_path[curr_depth]
  already exists and n.branch == RIGHT

  This function just sets the value
  to the right boolean value
  specified by 'assgn'


 ===================================*/
void ExQuant::make_node_assgn(int assgn)
{

  Node &n = Tree_path[curr_depth];
  assert(n.branch == RIGHT);
  assert(n.var == abs(assgn));

  if (assgn > 0) Var_state[n.var-1].val = 1;
  else Var_state[n.var-1].val = 0;



} /* end of function make_node_assgn */
