#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*==================================

  P R I N T _ D E S C R

  =================================*/
void ExQuant::print_descr()
{

  printf("pqe root [options]\n");
  printf("root specifies three files: root.cnf, root.txt and root.clauses\n");
  printf("options: ['d'mlrst] ['D'l3|l5|m|r|s]  [i|N|r|R|v] ['e'num] ['m'num] ['n' root]\n");
  printf("         ['p'|'pr'| 'pa'] ['S'seed] ['t'val]\n");
  printf("'d'          - set the decision making mode for quant. vars \n");
  printf("'D'          - set the decision making mode for non-quant. vars\n");
  printf("'e'num       - exit after 'num' Y-cubes have been generated\n");
  printf("'i'          - ignore the presence of 'wrong' Y-variables\n"); 
  printf("'G'gcnt      - switch on verbose after 'gcount' reaches value 'gcnt'\n");
  printf("'g' gcnt rng - ");
  printf("'m'num       - exit after 'num' D-sequents have been derived\n");
  printf("'n' root     - print the result into file 'root.cnf'\n");
  printf("'N'          - remove noise from the result\n");
  printf("'p'          - run preprocessing and stop\n");
  printf("              'pb' - run only BCP, 'pr' - run only identif. of redund. clauses,  'p' - run both\n");
  printf("'r'          - switch off checking if there is a signal Y-cube (the default is 'on')\n");
  printf("'R'          - just check if there is a signal Y-cube and exit\n");
  printf("'S'seed      - specify the seed to be used by random number generator\n");
  printf("'S'          - set the seed to time(0)\n");
  printf("'t'val       - sets the time limit (in seconts)\n");
  printf("'v'          - verbose output\n");
} /* end of funciton print_header */

/*===================================

  R E A D _ P A R A M E T E R S

  ===================================*/
void ExQuant::read_parameters(int argc,char *argv[])
{

  for (int i=2; i < argc; i++)
    switch (argv[i][0]) {   
    case 'd':
      if (argv[i][1] == 0) {
        printf("no choice of dec. mak. heur. for int. vars is given\n");
        exit(1);
      }
      switch(argv[i][1]) {
      case 'm':
        xvar_dec_mode = XVAR_MAX_ACT_CUBES;
        break;
      case 'l':
        xvar_dec_mode = XVAR_LIT_ACTIVITY;
        break;
      case 'r':
        xvar_dec_mode = XVAR_RAND;
        break;
      case 's':
        xvar_dec_mode = XVAR_TRIV;
        break;
      case 't':
        xvar_dec_mode = XVAR_MAX_ACT_SGNL_CUBES;
        break;
      default:
        p();
        printf("wrong dec. mak. heur. for X-vars\n");
        exit(1);
      }
      break;
    case 'D':
      if (argv[i][1] == 0) {
        p();
        printf("no choice of dec. mak. heur. for Y-vars is given\n");
        exit(1);
      }
      switch(argv[i][1]){
      case 'l':
        if (argv[i][2] == '3')
          yvar_dec_mode = YVAR_LIT_ACTIVITY3;
        else if (argv[i][2] == '5')
          yvar_dec_mode = YVAR_LIT_ACTIVITY3;
        else  {
          p();
          printf("wrong dec. mak. heur. for ext. vars\n");
          exit(1);
        }
        break; 
      case 'm':
        yvar_dec_mode = YVAR_MAX_ACT_CUBES;
        break;
      case 'r':
        yvar_dec_mode = YVAR_RAND;
        break;
      case 's':
        yvar_dec_mode = YVAR_TRIV;
        break;
      default:
        p();
        printf("wrong dec. mak. heur. for ext. vars\n");
        exit(1);
      } 
      break;  
    case 'e':
      max_num_ycubes = atoi(argv[i]+1);
      assert(max_num_ycubes >= 0);
      break;
    case 'G':
      start_gcount = read_int_num(argv[i]+1);
      assert(start_gcount >=0);
      break;
    case 'g':
      assert(i + 2 < argc);
      start_gcount = read_int_num(argv[i+1]);
      assert(start_gcount >= 0);
      range_gcount = read_int_num(argv[i+2]);
      assert(range_gcount > 0);
      i += 2;
      break;
    case 'i':
      ign_wrong_yvars = true;
      break;           
    case 'm':
      max_num_dseqs = read_int_num(argv[i]+1);
      if (max_num_dseqs <= 0) {
        p();
        printf("wrong maximal number of D-sequents (%s)\n",argv[i]+1);
        exit(1);
      }
      break;   
    case 'n':
      if (strlen(argv[i+1]) > MAX_NAME) {
        p();
        printf("too long file name %s\n",argv[i+1]);
        exit(1);
      }
      strcpy(res_root_name,argv[i+1]);
      i++;
      break;
    case 'N':
      noise = false;
      break;
    case 'p':
      prepr_mode = true;
      if (argv[i][1] == 0)  {
        prepr_bcp = true;
        prepr_red_cubes = true;
      }
      else if (argv[i][1] == 'b') prepr_bcp = true;
      else if (argv[i][1] == 'r') prepr_red_cubes = true;	
      else  {
        p();
        printf("wrong preprocessing mode %s\n",argv[i]);
        exit(1);
      }
      break;
    case 'r':
      rigor_check = false;
      break;
    case 'R':
      just_check_sgnl_cubes = true;
      break;   
    case 'S':
      if (strlen(argv[i]) > 1)  seed_value = atoi(argv[i]+1);
      else  {
        seed_value = time(0);
        printf("seed_value = %ld\n",seed_value);
      }
      break; 
    case 't':
      time_limit = atoi(argv[i]+1);
      if (time_limit <= 0) {
        p();
        printf("wrong time limit %s\n",argv[i]);
        exit(1);
      }
      break;
    case 'v':
      verbose = true;
      break;        
    default:
      p();
      printf("wrong key %c\n",argv[i][0]);
      exit(1);
    }
 
} /* end of function read_parameters */
