#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"



/*=================================================

  F I N D _ C A N D I D A T E _  R E D _ C U B E S

  This function returns the set of cubes
  'Res_cube_inds' such that the current active D-sequent 
  for cube 'cube_ind'  (where 'cube_ind' is in 
  'Res_cube_inds')  MAY  contain the branching
  variable x of the current node n

  ================================================*/
void ExQuant::find_candidate_red_cubes(CUBE &Red_cube_inds,CubeMap &Cube_inds)
{

  CubeMap::const_iterator i;
  for (i=Cube_inds.begin(); i != Cube_inds.end(); ++i)  {
    int cube_ind = i->first;
    if (Cube_info[cube_ind].state == ACTIVE) continue;
    Red_cube_inds.push_back(cube_ind);   
  }

} /* end of function find_candidate_red_cubes */


/*=============================================

  P R O C _ C U B E S _ S A T _ R B R A N C H

  This function is called to build
  D-sequents for cubes satisfied in 
  the right branch of 'n.var'.

  The D-sequents for these cubes have the 
  identical left part

  =============================================*/
void ExQuant:: proc_cubes_sat_rbranch(Node &n)
{

  int left_lit; 
  if (n.lval == 0) 
    left_lit = -n.var;
  else  // lval == 1
    left_lit = n.var;
  
  CUBE S;
  bool ok = quick_res(S,F[n.left_ecube],left_lit);
  assert(ok);
 
  CUBE &Cubes = New_sat_cubes[curr_depth];


  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (cube_ind == n.left_ecube) // the empty cube of the left branch
      continue;                   // is processed in a special way
    if (Cube_info[cube_ind].signal == 0)  {
      Cube_info[cube_ind].state = ACTIVE;
      continue;
    }
    num_resL_dseqs++;
    num_sc_dseqs++;
    num_fc_dseqs++;
    
    int L = max_cube_recovery_depth(S);
    set_new_act_dseq(cube_ind,S,false);
    check_red_cube_in(cube_ind,0,L,false);
  }
  
} /* end of function  proc_cubes_sat_rbranch */


