#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>
using namespace std;
#include "d0s_pqe.hh" 
#include "einline.hh"

/*===================================================

  R E C O V E R _ S A T _ M E D I A  _C U B E S

  This function sets the state of every media
  cube C   contained in 'New_sat_cubes[curr_depth]'
  to ACTIVE


  ======================================================*/
void  ExQuant::recover_sat_media_cubes()
{

  CUBE &Cubes = New_sat_cubes[curr_depth];
 
  CUBE Non_sgnl_cubes;
  for (uint i=0; i < Cubes.size(); i++)   { 
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].signal) continue;
    if (Cube_info[cube_ind].type == Y_CUBE) continue;
    assert(Cube_info[cube_ind].state == SATISFIED);
    recover_active_cube(cube_ind);
    Non_sgnl_cubes.push_back(cube_ind);	      
  }

  CUBE Red_cubes;

} /* end of function recover_sat_media_cubes */


/*========================================

  C H E C K _ S P E C _ C O N D

  This function returns 'true' if

  1) n.right_child_cube >= 0

  2) Cube_info[n.right_child_cube].signal == 1

  3) There is a cube that can
  be resolved with F[n.right_child_cube]
  on variable 'n.var' and this cube
  does not have the signal status


  ========================================*/
bool ExQuant::check_spec_cond(Node &n)
{

  int cube_ind = n.right_ecube;
 
  
  if (cube_ind < 0) return(false);

  if (Cube_info[cube_ind].signal == 0) return(false);

  
  CUBE *pCubes;

  if (n.lval == 0) pCubes = &Flits0[n.var-1];
  else pCubes = &Flits1[n.var-1];

  for (uint i=0; i < pCubes->size(); i++)    {
    int cube_ind1 = (*pCubes)[i];
    if (Cube_info[cube_ind1].signal) continue;
    if (distance(F[cube_ind],F[cube_ind1]) == 1)
      return(true);
  }  

  return(false);

} /* end of function check_spec_cond */

/*====================================

  R E V O K E _ S G N L _ S T A T U S

  This function 'revokes' the signal
  status for the cubes listed in
  'New_sgnl_cubes'

  ====================================*/
void ExQuant::revoke_sgnl_status()
{

  CUBE &Cubes = New_sgnl_cubes[curr_depth];
  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].signal == 0)	{
      p();
      printf("revoking the signal status of a media cube F[%d]\n",cube_ind);
      print_cube1(F[cube_ind],true);
      exit(1);
    }
    if (Cube_info[cube_ind].state == ACTIVE)	{
      num_act_sgnl_cubes--;
      decrease_sgnl_cube_scores(cube_ind);	 
    }
    else if (Cube_info[cube_ind].state == REDUNDANT)	{
      remove_lit_presence(cube_ind);
      Act_dseqs[cube_ind].clear(); 
      Cube_info[cube_ind].state = ACTIVE;
    }
     
    Cube_info[cube_ind].signal = 0;
    curr_num_sgnl_cubes--;
  }



}  /* end of function revoke_sgnl_status */

/*=========================================================

  C O N F I R M _ S G N L _ C U B E S

  This function update information regarding
  protecting cubes for the cubes listed 
  in Cubes.

  This function is almost identical to 'confirm_active_cubes'.
  Currently, the only difference is that 'confirm_sgnl_cubes'
  assumes every cube of 'Cubes' is a signal cube

  =========================================================*/
void ExQuant::confirm_sgnl_cubes(CUBE &Cubes,CUBE &Red_cubes,int depth)
{
  
  for (uint i=0; i < Cubes.size(); i++)   {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    int min_depth;
    int min_depth_lit;
     
  
    bool success = recover_protection(cube_ind,depth,min_depth_lit,min_depth);
      
      
    if (success) continue;

    CUBE S;
    	    

    seq_gen_red_cube1(S,min_depth_lit);
    num_frw_blocked_cube_dseqs++;

    set_new_act_dseq(cube_ind,S,false);
    check_red_cube_in(cube_ind,min_depth_lit,min_depth,true);
        
    Red_cubes.push_back(cube_ind);
  }   
	
} /* end of function confirm_sgnl_cubes */

/*=======================================

  S E T _ N E W _ S G N L  _ C U B E

  =========================================*/
void ExQuant::set_new_sgnl_cube(int cube_ind,int depth)
{

  Cube_info[cube_ind].signal = 1;
  if (depth >= 0)  New_sgnl_cubes[depth].push_back(cube_ind);
  Cube_info[cube_ind].depth = depth; 
  num_act_sgnl_cubes++;
  increase_sgnl_cube_scores(cube_ind);

} /* end of function set_new_sgnl_cube */

/*===================================

  A D D _ B C P _ S G N L _ C U B E S

  This function is called during BCP
  when an assignment 'assgn' is derived
  from a signal cube. This function gives
  the signal status to every active cube
  that 
  a) has the literal falsified by 'assgn'
  b) is not a signal cube yet

  ====================================*/
void ExQuant::add_bcp_sgnl_cubes(int assgn,int depth)
{

  CUBE *pCubes;
  if (assgn < 0) pCubes = &Flits0[-assgn-1];
  else pCubes = &Flits1[assgn-1];

  for (uint i=0; i < pCubes->size(); i++)   {
    int cube_ind = (*pCubes)[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].signal)   
      continue; // already a signal cube
    All_new_sgnl_cubes.push_back(cube_ind);
    set_new_sgnl_cube(cube_ind,depth);
    curr_num_sgnl_cubes++;
    if (curr_num_sgnl_cubes > max_num_sgnl_cubes)
      max_num_sgnl_cubes = curr_num_sgnl_cubes;
  }

} /* add_bcp_sgnl_cubes */
