#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"


/*=================================
  
  A D D _ A R R A Y

  ==================================*/
void add_array(CUBE &A,CUBE &B)
{

  for (uint i=0; i < B.size(); i++)
    A.push_back(B[i]);

} /* end of function add_array */

/*============================================

  C H E C K _ R E D _ C U B E _ O U T

  =============================================*/
void ExQuant::check_red_cube_out(int cube_ind)
{

  remove_lit_presence(cube_ind);
  Act_dseqs[cube_ind].clear(); 

} /* end of function check_red_cube_out */

/*=============================

  D I S T A N C E

  ============================*/
int  ExQuant::distance(CUBE &C1,CUBE &C2)
{

  hash_table_aux.change_marker();
  hash_table_aux.started_using();
  CUBE &table = hash_table_aux.table;
  int marker = hash_table_aux.marker;

  int count = 0;
 
  for (uint i=0; i < C1.size(); i++)
    if (C1[i] < 0) table[-C1[i]-1] = marker;
    else table[num_vars+ C1[i]-1] = marker;
    
 
  for (uint i=0; i < C2.size(); i++)  {
    int lit = C2[i];
    assert(abs(lit) <= num_vars);
    if (lit < 0) {
      if (table[-lit-1+num_vars] == marker) {
        count++;
        if (count >= 2) break;      	
      }   

    }
    else { 
      if (table[lit-1] == marker)  {
        count++;
        if (count >= 2) break; 
      }  
    }       
  }
  hash_table_aux.done_using();
  return(count);
}  /* end of function distance */

/*=====================================

  A C T _ S G N L _ C U B E S _ C N T

  =====================================*/
int ExQuant::act_sgnl_cubes_cnt(int var_ind)
{
  return(Act_sgnl_cubes_cnt0[var_ind]+Act_sgnl_cubes_cnt1[var_ind]);

} /* end of function act_sgnl_cubes_cnt */

/*=====================================

  T O T A L _ N U M _ R E S _ D S E Q S

  ======================================*/
long int  ExQuant::total_num_res_dseqs()
{
  return(num_resL_dseqs + num_resR_dseqs + num_resS_dseqs);
} /* end of function total_num_res_dseqs */

/*==============================================

  N U M _ B L O C K E D _ C U B E _ D S E Q S

  ===============================================*/
long int ExQuant::num_blocked_cube_dseqs()
{
  return(num_frw_blocked_cube_dseqs + num_bkw_blocked_cube_dseqs);
} /*end of function num_blocked_cube_dseqs */


/*===================

  Y V A R _ N O D E

  ==================*/
bool ExQuant::yvar_node(int depth)
{

  Node &n = Tree_path[depth];
  return(Var_state[n.var-1].xvar == 0);

} /* end of function yvar_ node */

/*==========================================

  R E C O V E R _ A C T I V E _ C U B E

  ===========================================*/
void ExQuant::recover_active_cube(int cube_ind)
{

  assert(cube_ind < (int) F.size());
  assert(cube_ind < (int) Act_dseqs.size());
  Act_dseqs[cube_ind].clear();
  if (Cube_info[cube_ind].state != ACTIVE)   { 
    if (Cube_info[cube_ind].signal)   {
      num_act_sgnl_cubes++;
      increase_sgnl_cube_scores(cube_ind);
    }
    Cube_info[cube_ind].state = ACTIVE;
  }
    
} /* end of function recover_active_cube */

/*=================================================

  C H E C K _ S A T _ C U B E _ I N

  =================================================*/
void ExQuant::check_sat_cube_in(int cube_ind,int assgn)
{
 
  if (Cube_info[cube_ind].state == ACTIVE) 
    if (Cube_info[cube_ind].signal)      {
      num_act_sgnl_cubes--;
      decrease_sgnl_cube_scores(cube_ind);
    }
  Cube_info[cube_ind].state = SATISFIED;
  Cause_lit[cube_ind] = assgn;
  New_sat_cubes[curr_depth].push_back(cube_ind); 
  
  Red_cube_depth[cube_ind] = curr_depth;
 
     
} /* end of function check_sat_cube_in */



/*========================================

  C H E C K _ R E D _ C U B E _ I N

  ==========================================*/
void ExQuant::check_red_cube_in(int  cube_ind,int lit,int rec_depth,
                                bool recompute_scores)
{
 
  assert((curr_depth == -1) || (Cube_info[cube_ind].signal == 1)); 
 
  if (Cube_info[cube_ind].state == ACTIVE) 
    if (Cube_info[cube_ind].signal)    {
      num_act_sgnl_cubes--;
      if (recompute_scores)
	decrease_sgnl_cube_scores(cube_ind);
    }
  Cube_info[cube_ind].state = REDUNDANT;
  Cause_lit[cube_ind] = lit;
  Red_cube_depth[cube_ind] = rec_depth;
  
 
  
} /* check_red_cube_in */


