#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

using namespace std;
#include "d0s_pqe.hh"
#include "einline.hh"

/*========================================

     Y V A R _ A R R _ S Q Z _ C O N D

  =======================================*/
inline bool ExQuant::yvar_arr_sqz_cond()
{

  return ((Sgnl_yvars.size() > 20) && 
	  ((int) Sgnl_yvars.size() > 2*num_sgnl_yvars));
   

} /* end of function yvar_arr_sqz_cond */


/*============================================

  R E M _ Y V A R _ F R O M _ A R R A Y

 =============================================*/
void ExQuant::rem_yvar_from_array(int var_ind)
{

  if (Var_state[var_ind].in_array == 0)    {
    p();
    printf("var = %d\n",var_ind+1);
    printf("gcount = %ld\n",gcount);
    exit(1);
  }
  
  if (yvar_arr_sqz_cond())
    squeeze_vars(Sgnl_yvars);


} /* end of function rem_yvar_from_array */

/*============================================

     A D D _ Y V A R _ T O _ A R R A Y

  ===========================================*/
void ExQuant::add_yvar_to_array(int var_ind)
{

  if (Var_state[var_ind].in_array == 0)   {
    Sgnl_yvars.push_back(var_ind+1);
    Var_state[var_ind].in_array = 1;
  }

  if (yvar_arr_sqz_cond())
    squeeze_vars(Sgnl_yvars);

} /* end of function add_yvar_to_array */

/*=================================

    S Q U E E Z E _ V A R S

 ==================================*/
void ExQuant::squeeze_vars(CUBE &Vars)
{

  int count = 0;
  for (uint i=0; i < Vars.size(); i++)    {
    int var_ind = Vars[i]-1;
    if (act_sgnl_cubes_cnt(var_ind) == 0)	{
      count++;
      Var_state[var_ind].in_array = 0;
      continue;
    }
    if (count > 0)
      Vars[i-count] = Vars[i];
  }

  Vars.resize(Vars.size()-count);

}  /* end of function squeeze_free_vars */

/*========================================

     I N T _ A R R _ S Q Z _ C O N D

  =======================================*/
inline bool ExQuant::int_arr_sqz_cond()
{

  return ((Sgnl_xvars.size() > 20) && 
	  ((int) Sgnl_xvars.size() > 2*num_sgnl_xvars));
   

} /* end of function arr_sqz_cond */


/*============================================

  R E M _ X V A R _ F R O M _ A R R A Y

 =============================================*/
void ExQuant::rem_xvar_from_array()
{

  
  if (int_arr_sqz_cond())
    squeeze_vars(Sgnl_xvars);


} /* end of function rem_xvar_from_array */

/*==================================================

  A D D _ X V A R _ T O _ A R R A Y

  ===================================================*/
void ExQuant::add_xvar_to_array(int var_ind)
{

  
  if (Var_state[var_ind].in_array == 0)  {
    Sgnl_xvars.push_back(var_ind+1);
    Var_state[var_ind].in_array = 1;
  }

  if (int_arr_sqz_cond())
    squeeze_vars(Sgnl_xvars);

} /* end of function add_xvar_to_array */



/*====================================================

  I N C R E A S E _ S G N L _ C U B E _ S C O R E S

  ===================================================*/
void ExQuant::increase_sgnl_cube_scores(int cube_ind)
{

  CUBE &C = F[cube_ind];
 

  for (uint i=0; i < C.size(); i++)  {
    int var_ind = abs(C[i])-1;
    if (C[i] < 0) Act_sgnl_cubes_cnt0[var_ind]++;
    else Act_sgnl_cubes_cnt1[var_ind]++;

    if (act_sgnl_cubes_cnt(var_ind) == 1) { 
      if (Var_state[var_ind].xvar)    {
        num_sgnl_xvars++;
        add_xvar_to_array(var_ind);
      }
      else {// a Y-variable  	  	   
        num_sgnl_yvars++;
        add_yvar_to_array(var_ind);
      }
    }
  }


} /* end of function increase_sgnl_cube_scores */

/*=======================================================

   D E C R E A S E _ S G N L _ C U B E _ S C O R E S

  ========================================================*/
void ExQuant::decrease_sgnl_cube_scores(int cube_ind)
{
  CUBE &C = F[cube_ind];

  for (uint i=0; i < C.size(); i++)    {
    int var_ind = abs(C[i])-1;
    if (act_sgnl_cubes_cnt(var_ind) <= 0) 	{
      p();
      printf(" act_sgnl_cubes_cnt[%d] == %d\n",var_ind,act_sgnl_cubes_cnt(var_ind));
      printf("Act_sgnl_cubes_cnt0(%d)=%d\n",var_ind,Act_sgnl_cubes_cnt0[var_ind]);
      printf("Act_sgnl_cubes_cnt1(%d)=%d\n",var_ind,Act_sgnl_cubes_cnt1[var_ind]);
      printf("decreasing scores due to cube number %d -> ",cube_ind); cout << C << endl;
      print_cube_state(cube_ind);
      printf("num_act_sgnl_cubes=%d\n",num_act_sgnl_cubes);
      if (Cube_info[cube_ind].state == REDUNDANT)
        print_dseq3(cube_ind);
      print_sgnl_status(cube_ind);
      printf("gcount = %ld\n",gcount);
      print_alg_state(false);
      printf("F.size() = %d\n",(int) F.size());
      exit(1);
    }
     
    if (C[i] < 0) Act_sgnl_cubes_cnt0[var_ind]--;
    else Act_sgnl_cubes_cnt1[var_ind]--;

    if (act_sgnl_cubes_cnt(var_ind) == 0) { 
      if (Var_state[var_ind].xvar) 	  {
        num_sgnl_xvars--;
        rem_xvar_from_array();
      }
      else {// a Y-variable
        num_sgnl_yvars--;
        rem_yvar_from_array(var_ind);
      }	
    }
  }

} /* end of function decrease_sgnl_cube_scores */
 
